/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: audio_oss_out.c,v 1.23 2001/04/29 22:05:30 siggi Exp $
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>
#if defined(__OpenBSD__)
#include <soundcard.h>
#elif defined(__FreeBSD__)
#include <machine/soundcard.h>
#else
#include <sys/soundcard.h>
#endif
#include <sys/ioctl.h>
#include <inttypes.h>

#include "xine.h"
#include "monitor.h"
#include "audio_out.h"
#include "resample.h"
#include "metronom.h"
#include "ac3.h"
#include "utils.h"

#define AUDIO_NUM_FRAGMENTS     15
#define AUDIO_FRAGMENT_SIZE   8192

#define GAP_TOLERANCE        15000
#define MAX_MASTER_CLOCK_DIV  5000

extern uint32_t xine_debug;

extern int use_realtime_sync;
extern int force_realtime_sync;
static int checked_getoptr;

#define DSP_TEMPLATE "/dev/dsp%d"

typedef struct _audio_oss_globals {

  uint32_t       bytes_in_buffer;      /* total btyes written to audio dev           */
  uint32_t       sync_bytes_in_buffer; /* for vpts <-> samplecount assoc             */
  uint32_t       sync_vpts;            /* this syncpoint is used as a starting point */
  int32_t        bytes_per_kpts;       /* bytes per 1024/90000 sec                   */

  uint32_t       last_vpts;            /* vpts at which last package ends            */
  uint32_t       output_sample_rate;
  uint32_t       input_sample_rate;
  uint32_t       num_channels;

  int            audio_started;
  int            audio_has_realtime;    /* no need for soft audio sync               */
  int            audio_fd;
  int32_t        output_rate_correction;    /* not used -- leave to pad cache line   */

  int            mode;
  int            audio_step;           /* pts per 32 768 samples (sample = #bytes/2) */
  int16_t       *zero_space;

  char           audio_dev[20];

} audio_oss_globals_t;

static audio_oss_globals_t gAudioOSS;

/*
 * open the audio device for writing to
 */
static int ao_open(uint32_t bits, uint32_t rate, int mode)
{
  int tmp;
  int fsize;

  if ((mode != AO_MODE_STEREO) && (mode != AO_MODE_MONO)) {
    printf ("audio_oss_out: OSS Driver only supports mono/stereo output modes at the moment\n");
    return -1;
  }

  if (gAudioOSS.audio_fd > -1) {

    if (rate == gAudioOSS.input_sample_rate && mode == gAudioOSS.mode)
      return 1;

    close (gAudioOSS.audio_fd);
  }

  printf ("audio_oss_out: ao_open rate=%d, mode=%d, bps=%d\n", rate, mode, bits);

  gAudioOSS.bytes_in_buffer        = 0;
  gAudioOSS.sync_bytes_in_buffer   = 0;
  gAudioOSS.sync_vpts              = 0;
  gAudioOSS.bytes_per_kpts         = 0;

  gAudioOSS.last_vpts              = 0;
  gAudioOSS.output_sample_rate     = 0;
  gAudioOSS.input_sample_rate      = rate;
  gAudioOSS.num_channels           = 0;
  gAudioOSS.audio_started          = 0;
  gAudioOSS.output_rate_correction = 0;

  /*
   * open audio device
   */

  gAudioOSS.audio_fd=open(gAudioOSS.audio_dev,O_WRONLY|O_NDELAY);
  if(gAudioOSS.audio_fd < 0) {
    printf("audio_oss_out: Opening audio device %s: %s\n",
	   gAudioOSS.audio_dev, strerror(errno));
    return -1;
  }
  
  /* We wanted non blocking open but now put it back to normal */
  fcntl(gAudioOSS.audio_fd, F_SETFL, fcntl(gAudioOSS.audio_fd, F_GETFL)&~FNDELAY);

  /*
   * configure audio device
   */

  gAudioOSS.mode = mode;
  tmp = (mode == AO_MODE_STEREO) ? 1 : 0;
  ioctl(gAudioOSS.audio_fd,SNDCTL_DSP_STEREO,&tmp);

  gAudioOSS.num_channels = tmp+1;
  xprintf (VERBOSE|AUDIO, "audio_oss_out: %d channels\n",gAudioOSS.num_channels);

  /*
   * set sample size, NOTE - subsequent code assumes 16bit samples!
   */
  tmp = bits;
  ioctl(gAudioOSS.audio_fd,SNDCTL_DSP_SAMPLESIZE,&tmp);

  /*
   * Previous calls don't necessarily set endiness the way
   * we need. On powermac with AWACS dmasound driver they
   * switch us from BE to LE, which is not what we want!
   */
  tmp = bits == 16 ? AFMT_S16_NE : AFMT_S8;
  ioctl(gAudioOSS.audio_fd,SNDCTL_DSP_SETFMT,&tmp);

  tmp = gAudioOSS.input_sample_rate;
#ifdef FORCE_44K_MAX
  if(tmp > 44100)
     tmp = 44100;
#endif
  ioctl(gAudioOSS.audio_fd,SNDCTL_DSP_SPEED, &tmp);
  gAudioOSS.output_sample_rate = tmp;

  xprintf (VERBOSE|AUDIO, 
     "audio_oss_out: audio rate : %d requested, %d provided by device/sec\n",
     gAudioOSS.input_sample_rate, gAudioOSS.output_sample_rate);

  gAudioOSS.audio_step         = (uint32_t) 90000 * (uint32_t) 32768 
                                 / gAudioOSS.input_sample_rate;

  gAudioOSS.bytes_per_kpts     = gAudioOSS.output_sample_rate * gAudioOSS.num_channels * 2 * 1024 / 90000;

  xprintf (VERBOSE|AUDIO, "audio_oss_out: %d bytes/kpts, \n", gAudioOSS.bytes_per_kpts);
  xprintf (VERBOSE|AUDIO, "audio_oss_out : audio_step %d pts per 32768 samples\n", gAudioOSS.audio_step);

  metronom_set_audio_rate (gAudioOSS.audio_step);

  /*
   * audio buffer size handling
   */

  tmp=0 ;
  fsize = AUDIO_FRAGMENT_SIZE;
  while (fsize>0) {
    fsize /=2;
    tmp++;
  }
  tmp--;

  tmp = (AUDIO_NUM_FRAGMENTS << 16) | tmp ;

  xprintf (VERBOSE|AUDIO, 
             "audio_oss_out: Audio buffer fragment info : %x\n",tmp);

  ioctl(gAudioOSS.audio_fd,SNDCTL_DSP_SETFRAGMENT,&tmp); 

  /*
   * Final check of realtime capability, make sure GETOPTR
   * doesn't return an error.
   */
  if ( gAudioOSS.audio_has_realtime && !checked_getoptr ) {
    count_info info;
    int ret = ioctl(gAudioOSS.audio_fd, SNDCTL_DSP_GETOPTR, &info);
    if ( ret == -1 && errno == EINVAL ) {
      gAudioOSS.audio_has_realtime = 0;
      printf("audio_oss_out: Audio driver SNDCTL_DSP_GETOPTR reports %s,"
               " disabling realtime sync...\n", strerror(errno) );
      printf("audio_oss_out: ...Will use video master clock for soft-sync instead\n");
      printf("audio_oss_out: ...There may be audio/video synchronization issues\n");
    }
    checked_getoptr = 1;
  }

  return 1;
}

static uint32_t ao_get_current_vpts (void) {

  int      pos ;
  int32_t  diff ;
  uint32_t vpts ;
  
  count_info info;
  
  if (gAudioOSS.audio_started) {
    ioctl (gAudioOSS.audio_fd, SNDCTL_DSP_GETOPTR, &info);

    pos = info.bytes;

  } else
    pos = 0;

  diff = gAudioOSS.sync_bytes_in_buffer - pos;
  
  vpts = gAudioOSS.sync_vpts - diff * 1024 / gAudioOSS.bytes_per_kpts;

  xprintf (AUDIO|VERBOSE,"audio_oss_out: get_current_vpts pos=%d diff=%d vpts=%d sync_vpts=%d\n",
	   pos, diff, vpts, gAudioOSS.sync_vpts);

  return vpts;
}

static void ao_fill_gap (uint32_t pts_len) {

  int num_bytes = pts_len * gAudioOSS.bytes_per_kpts / 1024;

  num_bytes &= ~3;

  printf ("audio_oss_out: inserting %d 0-bytes to fill a gap of %d pts\n",num_bytes, pts_len);
  
  gAudioOSS.bytes_in_buffer += num_bytes;
  gAudioOSS.last_vpts += pts_len;
  
  if (!gAudioOSS.audio_started) {
     metronom_resume_clock();
     gAudioOSS.audio_started = 1;
  }

  while (num_bytes>0) {
    if (num_bytes>8192) {
      write(gAudioOSS.audio_fd, gAudioOSS.zero_space, 8192);
      num_bytes -= 8192;
    } else {
      write(gAudioOSS.audio_fd, gAudioOSS.zero_space, num_bytes);
      num_bytes = 0;
    }
  }

}

static void ao_write_audio_data(int16_t* output_samples, uint32_t num_samples, 
				uint32_t pts_)
{

  uint32_t vpts,
           audio_vpts,
           master_vpts;
  int32_t  diff, gap;
  int      bDropPackage;
  uint16_t sample_buffer[8192];

  if (gAudioOSS.audio_fd<0)
    return;

  vpts        = metronom_got_audio_samples (pts_, num_samples);

  xprintf (VERBOSE|AUDIO, 
     "audio_oss_out: got %d samples, pts_=%d vpts=%d, last_vpts=%d\n",
	   num_samples, pts_, vpts, gAudioOSS.last_vpts);

  if (!gAudioOSS.audio_started) {
     /* see comment before write */
     metronom_stop_clock();
  }

  bDropPackage = 0;
  if ( gAudioOSS.audio_has_realtime || !gAudioOSS.audio_started ) {

  /*
   * check if these samples "fit" in the audio output buffer
   * or do we have an audio "gap" here?
   *
   * if vpts wraps around zero, gap will need to be
   * adjusted -- call me paranoid, but it's happened,
   * though probably due to buffer curruption problems. 
   */
  
  gap = vpts - gAudioOSS.last_vpts ;
  
  if ( (vpts ^ gAudioOSS.last_vpts) & 0x8000000 ) {
    uint32_t diff2 = (uint32_t)-1 - (uint32_t)gap;
    /* straddling zero or 0x8000000 */
    printf("audio_oss_out: *** vpts counters wrapping around zero: "
              "vpts=%u last_vpts=%u gap=%d diff2=%d ***\n",
         vpts, gAudioOSS.last_vpts, gap, diff2);
    if ( diff2 < (uint32_t)gap ) {
       /* diff2 is abs gap across zero */
       gap = vpts & 0x80000000 ? /* vpts trails ? */
               -(int32_t)diff2 : (int32_t)diff2;
      printf("audio_oss_out: *** vpts counters wrapping around zero: vpts=%u last_vpts=%u gap=%d ***\n",
         vpts, gAudioOSS.last_vpts, gap);
    }
  }

  /*
    printf ("audio_oss_out: gap = %d - %d + %d = %d\n",
    vpts, gAudioOSS.last_vpts, diff, gap);
  */

  
  if (gap>GAP_TOLERANCE) {
    printf ("audio_oss_out: filling gap of %d: vpts=%d, last_vpts=%d\n",
              gap, vpts, gAudioOSS.last_vpts);
    ao_fill_gap (gap);
    printf ("audio_oss_out: filled gap of %d: vpts=%d, last_vpts=%d\n",
              gap, vpts, gAudioOSS.last_vpts);
  } else if (gap<-GAP_TOLERANCE) {
    bDropPackage = 1;
  }


    /*
     * sync on master clock
     */

    audio_vpts  = ao_get_current_vpts () ;
    master_vpts = metronom_get_current_time ();
    diff        = audio_vpts - master_vpts;

    /*
     * wrap of one vpts counter around zero,
     * when otherwise very close, will cause large diff!
     */
    if ( (audio_vpts ^ master_vpts) & 0x80000000 ) {
      /* stradling zero or 0x8000... */
      uint32_t diff2 = (uint32_t)-1 - (uint32_t)diff;
      if ( diff2 < (uint32_t)diff ) {
        /* stradling zero, diff2 is abs difference */
        diff = audio_vpts & 0x80000000 ?   /* audio_vpts trails? */
                  -(int32_t)diff2 : (int32_t)diff2;
        printf("audio_oss_out: *** vpts counters wrapping around zero: audio_vpts=%u master_vpts=%u diff=%d ***\n",
           audio_vpts, master_vpts, diff);
      }
    }

    xprintf (AUDIO|VERBOSE, "audio_oss_out: syncing on master clock: audio_vpts=%d master_vpts=%d\n",
	   audio_vpts, master_vpts);
    /*
    printf ("audio_oss_out: audio_vpts=%d <=> master_vpts=%d (diff=%d)\n",
  	  audio_vpts, master_vpts, diff);
    */
    /*
     * method 1 : resampling
     */

    /*
    if (abs(diff)>5000) {

      if (diff>5000) {
        ao_fill_gap (diff);
      } else if (diff<-5000) {
        bDropPackage = 1;
      }

    } else if (abs(diff)>1000) {
      gAudioOSS.output_rate_correction = diff/10 ; 
     
      printf ("audio_oss_out: diff = %d => rate correction : %d\n", 
               diff, gAudioOSS.output_rate_correction);  
    
      if ( gAudioOSS.output_rate_correction < -500)
        gAudioOSS.output_rate_correction = -500;
      else if ( gAudioOSS.output_rate_correction > 500)
        gAudioOSS.output_rate_correction = 500;
    }
    */

    /*
     * method 2: adjust master clock
     */
  
  
    if (abs(diff)>MAX_MASTER_CLOCK_DIV) {
      printf ("audio_oss_out: master clock adjust time %u -> %u (diff: %d)\n", master_vpts, audio_vpts, diff); 
      metronom_adjust_clock (audio_vpts); 
    }
  
  } /* has realtime */

  /*
   * resample and output samples
   */

  if (!bDropPackage) {
    int num_output_samples = num_samples 
                               * (gAudioOSS.output_sample_rate
                                   /* + gAudioOSS.output_rate_correction */) 
                               / gAudioOSS.input_sample_rate;
    int num_output_bytes =  num_output_samples * 2 * gAudioOSS.num_channels;

    audio_out_resample_stereo (output_samples, num_samples,
			       sample_buffer, num_output_samples);
   
    if (!gAudioOSS.audio_started) {
       /*
        *  There is enough startup overhead to seriously throw
        *  off sync for the first sample -- especially with
        *  resampling. Stopping or adjusting the clock here
        *  seems to really help.
        */
       xprintf (AUDIO|VERBOSE, "audio_oss_out: first audio packet ending "
                 "pts: %u, current vpts: %u, sync_vpts: %u\n", 
                 pts_, vpts, gAudioOSS.sync_vpts);
       metronom_resume_clock ();
       gAudioOSS.audio_started    = 1;
    }

    write(gAudioOSS.audio_fd, sample_buffer, num_output_bytes);

    xprintf (AUDIO|VERBOSE, "audio_oss_out: audio package written\n");
    
    /*
     * remember vpts
     */
    
    gAudioOSS.sync_vpts            = vpts;
    gAudioOSS.sync_bytes_in_buffer = gAudioOSS.bytes_in_buffer;

    /*
     * step values
     */
    
    gAudioOSS.bytes_in_buffer += num_output_bytes;
  } else {
    if (!gAudioOSS.audio_started) 
       metronom_resume_clock();
    gAudioOSS.sync_vpts            = vpts;
    printf ("audio_oss_out: audio package (vpts = %d) dropped\n", vpts);
  }
  
  gAudioOSS.last_vpts        = vpts + num_samples * 90000 
                                       / gAudioOSS.input_sample_rate ; 
}


static void ao_close(void)
{
  close(gAudioOSS.audio_fd);
  gAudioOSS.audio_fd = -1;
}

static int ao_is_mode_supported (int mode) {
  return ((mode == AO_MODE_STEREO) || (mode == AO_MODE_MONO));
}

static ao_functions_t audio_ossout = {
  ao_is_mode_supported,
  ao_open,
  ao_write_audio_data,
  ao_close,
};


ao_functions_t *audio_ossout_init (void) 
{
  int   caps;
  char  devname[] = "/dev/dsp\0\0\0";
  int   best_rate;
  int   rate ;
  int   devnum;
  int   audio_fd;

  /*
   * find best device driver/channel
   */

  xprintf (VERBOSE|AUDIO, "audio_oss_out: Opening audio device...");
  devnum = 0;
  best_rate = 0;
  while (devnum<16) {
    audio_fd=open(devname,O_WRONLY|O_NDELAY);

    if (audio_fd>0) {

      xprintf (VERBOSE|AUDIO, 
                "audio_oss_out: testing %s\n", gAudioOSS.audio_dev);

      /* test bitrate capability */
      
      rate = 48000;
      ioctl(audio_fd,SNDCTL_DSP_SPEED, &rate);
      if (rate>best_rate) {
	strncpy (gAudioOSS.audio_dev, devname, 19);
	best_rate = rate;
      }
      
      close (audio_fd);
    }
    else {
      xprintf(VERBOSE|AUDIO, "audio_oss_out: %s - opening audio device %s.\n",
              strerror(errno), gAudioOSS.audio_dev);
    }

    sprintf(devname, DSP_TEMPLATE, devnum);
    devnum++;
  }

  /*
   * open that device
   */

  audio_fd=open(gAudioOSS.audio_dev, O_WRONLY|O_NDELAY);

  if(audio_fd < 0) 
  {
    xprintf(VERBOSE|AUDIO, "audio_oss_out: %s - opening audio device %s\n",
	   strerror(errno), gAudioOSS.audio_dev);

    return NULL;

  } else
    xprintf (VERBOSE|AUDIO, "audio_oss_out: %s\n", gAudioOSS.audio_dev);

  ioctl (audio_fd, SNDCTL_DSP_GETCAPS, &caps);

  if ( (caps & DSP_CAP_REALTIME) > 0) {
    xprintf (VERBOSE|AUDIO, "audio_oss_out: realtime check: passed :-)\n");
    if( use_realtime_sync )
       gAudioOSS.audio_has_realtime = 1;
    else
       gAudioOSS.audio_has_realtime = 0;
  } else {
    printf ("audio_out_oss: realtime check: *FAILED* :-(((((\n");
    gAudioOSS.audio_has_realtime = 0;
  }

  if( !gAudioOSS.audio_has_realtime ) {
    if( !force_realtime_sync ) {
      printf("audio_oss_out: Audio driver realtime sync disabled...\n");
      printf("audio_oss_out: ...Will use video master clock for soft-sync instead\n");
      printf("audio_oss_out: ...There may be audio/video synchronization issues\n");
    }
    else {
      gAudioOSS.audio_has_realtime = 1;
      printf("audio_oss_out: Force realtime sync specified...\n");
      printf("audio_oss_out: ...Will attempt SNDCTL_DSP_GETOPTR.\n");
    }
  }

  /*
   * Not used anymore
   *
  if ((caps & DSP_CAP_TRIGGER) > 0) {
    xprintf (VERBOSE|AUDIO, "audio_oss_out: trigger check : passed :-)\n");
  } else {
    xprintf (VERBOSE|AUDIO, "audio_oss_out: trigger check : *FAILED* :-(((((\n");
  }
  *
  */

  close (audio_fd);

  gAudioOSS.output_sample_rate = 0;

  gAudioOSS.zero_space = xmalloc (8192);

  return &audio_ossout;
}
