/*
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: buffer.c,v 1.11 2001/04/29 22:05:29 siggi Exp $
 *
 *
 * contents:
 *
 * buffer_entry structure - serves as a transport encapsulation
 *   of the mpeg audio/video data through xine
 *
 * free buffer pool management routines
 *
 * FIFO buffer structures/routines
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include "buffer.h"
#include "utils.h"

#define STABLE 

/*
 * global variables
 */

buf_element_t   *gBufferPoolTop;    /* a stack actually */
pthread_mutex_t  gBufferPoolMutex;
pthread_cond_t   gBufferPoolCondNotEmpty;
int              gBufferPoolNumFree;

/*
 * put a previously allocated buffer element back into the buffer pool
 */
static void buffer_pool_free (buf_element_t *pBufElement) {

  pthread_mutex_lock (&gBufferPoolMutex);

#if defined(DEBUG) || !defined(STABLE)
    if ( gBufferPoolNumFree < 0 
         || (gBufferPoolNumFree <= 0 && gBufferPoolTop)
         || (gBufferPoolNumFree != 0 && !gBufferPoolTop) ) {
       printf( "buffer_pool_free: *** buffer corrupt - num=%d, top=0x%lx ***\n",
               gBufferPoolNumFree, (long)gBufferPoolTop );
       gBufferPoolTop = NULL;
       gBufferPoolNumFree = 0;
    }
#endif

  pBufElement->next = gBufferPoolTop;
  gBufferPoolTop = pBufElement;

  gBufferPoolNumFree++;

  pthread_cond_signal (&gBufferPoolCondNotEmpty);

  pthread_mutex_unlock (&gBufferPoolMutex);
}

static void *malloc_aligned (int size, int alignment) {
  void *pMem;
  int  align;

  pMem = xmalloc (size+alignment);

#if 0
  while ((unsigned long) pMem % alignment)
    pMem++;
#else
  align = (unsigned long)pMem % alignment;
  if( align )
    pMem += alignment - align;
#endif

  return pMem;
}

/*
 * check if there are no more free elements
 */
static int buffer_pool_isempty (void) {
  return gBufferPoolNumFree<7;
}

/*
 * check if there are no more free elements
 */
static buf_element_t *buffer_pool_alloc (void) {
  
  buf_element_t *pBuf;

  pthread_mutex_lock (&gBufferPoolMutex);
  
#if defined(DEBUG) || !defined(STABLE)
   if ( (!gBufferPoolTop && gBufferPoolNumFree != 0)
        || (gBufferPoolTop && gBufferPoolNumFree <= 0) ) {
      printf( "buffer_pool_alloc: *** buffer corrupt - num=%d, top=0x%lx ***\n",
               gBufferPoolNumFree, (long)gBufferPoolTop );
      gBufferPoolTop = NULL;
      gBufferPoolNumFree = 0;
      pthread_cond_init (&gBufferPoolCondNotEmpty,NULL);
    }
#endif

  while (!gBufferPoolTop) {
    pthread_cond_wait (&gBufferPoolCondNotEmpty, &gBufferPoolMutex);
  }

  pBuf = gBufferPoolTop;
  gBufferPoolTop = gBufferPoolTop->next;
  gBufferPoolNumFree--;

  if ( gBufferPoolNumFree <= 0 || !gBufferPoolTop ) {
#if defined(DEBUG) || !defined(STABLE)
    if ( gBufferPoolNumFree != 0 || gBufferPoolTop ) {
       printf( "buffer_pool_alloc: *** buffer corrupt - num=%d, top=0x%lx ***\n",
               gBufferPoolNumFree, (long)gBufferPoolTop );
    }
#endif
    gBufferPoolTop = NULL;
    gBufferPoolNumFree = 0;
    pthread_cond_init (&gBufferPoolCondNotEmpty,NULL);
  }

  pthread_mutex_unlock (&gBufferPoolMutex);

  return pBuf;
}

/*
 * allocate and initialize new (empty) fifo buffer
 */
static fifo_buffer_t *fifo_buffer_new (void) {

  fifo_buffer_t *pFifo;

  pFifo = xmalloc (sizeof (fifo_buffer_t));

  pFifo->mpFirst = NULL;
  pFifo->mpLast  = NULL;

  pthread_mutex_init (&pFifo->mMutex, NULL);
  pthread_cond_init (&pFifo->mNotEmpty, NULL);

  return pFifo;
}

/*
 * append buffer element to fifo buffer
 */
static void fifo_buffer_put (fifo_buffer_t *pFifo, buf_element_t *pBufElement) {
  
  pthread_mutex_lock (&pFifo->mMutex);

  pBufElement->next = NULL;
  if (pFifo->mpLast) 
    pFifo->mpLast->next = pBufElement;
  else {
#if defined(DEBUG) || !defined(STABLE)
    if( pFifo->mpFirst )
       printf( "fifo_buffer_put: *** fifo corrupt - pFifo->mpFirst=0x%lx,"
                 " pFifo->mpLast=0x%lx ***\n",
                (long)pFifo->mpFirst, (long)pFifo->mpLast ); 
#endif
    pFifo->mpFirst = pBufElement;
  }
  pFifo->mpLast  = pBufElement;

  pthread_cond_signal (&pFifo->mNotEmpty);

  pthread_mutex_unlock (&pFifo->mMutex);
}

/*
 * get element from fifo buffer
 */
static buf_element_t *fifo_buffer_get (fifo_buffer_t *pFifo) {

  buf_element_t *pBuf;
  
  pthread_mutex_lock (&pFifo->mMutex);

  while (pFifo->mpFirst==NULL) {
    pthread_cond_wait (&pFifo->mNotEmpty, &pFifo->mMutex);
  }

  pBuf = pFifo->mpFirst;

  pFifo->mpFirst = pFifo->mpFirst->next;
  if ( pFifo->mpFirst == NULL ) {
#if defined(DEBUG) || !defined(STABLE)
    if( pBuf != pFifo->mpLast || pBuf->next )
       printf( "fifo_buffer_get: *** fifo corrupt - pFifo->mpFirst=0x%lx,"
                 " pFifo->mpFirst->next=ox%lx, pFifo->mpLast=0x%lx ***\n",
                (long)pBuf, (long)pBuf->next, (long)pFifo->mpLast ); 
#endif
    pFifo->mpLast = NULL;
    pthread_cond_init (&pFifo->mNotEmpty, NULL);
  }
  pBuf->next = NULL;

  pthread_mutex_unlock (&pFifo->mMutex);

  return pBuf;
}

/*
 * buffer status
 */
static int fifo_buffer_is_empty (fifo_buffer_t *pFifo) {
  return (pFifo->mpFirst == NULL);
}

/*
 * clear buffer (put all contained buffer elements back into buffer pool)
 */
static void fifo_buffer_clear (fifo_buffer_t *pFifo) {
  
  buf_element_t *pBuf;

  pthread_mutex_lock (&pFifo->mMutex);

  while (pFifo->mpFirst != NULL) {

    pBuf = pFifo->mpFirst;

    pFifo->mpFirst = pFifo->mpFirst->next;
    if (pFifo->mpFirst==NULL) {
      pFifo->mpLast = NULL;
      pthread_cond_init (&pFifo->mNotEmpty, NULL);
    }
    pBuf->next = NULL;
    buffer_pool_free (pBuf);
  }

  pthread_mutex_unlock (&pFifo->mMutex);
}

static fifobuf_functions_t fifobuf_op = {
  buffer_pool_isempty,
  buffer_pool_alloc,
  buffer_pool_free,
  fifo_buffer_new,
  fifo_buffer_put,
  fifo_buffer_get,
  fifo_buffer_is_empty,
  fifo_buffer_clear
};

/*
 * init buffer pool, allocate nNumBuffers of buf_size bytes each 
 */
fifobuf_functions_t *buffer_pool_init (int nNumBuffers, uint32_t buf_size) {

  int i;
  const int alignment = 2048;
  char *pMultiBuffer = NULL;

  if ((buf_size % alignment) == 0) {
    printf ("Allocating %d buffers of %ld bytes in one chunk (alignment = %d)\n", nNumBuffers, (long int)buf_size, alignment);
      pMultiBuffer = malloc_aligned (nNumBuffers * buf_size, alignment);
  }

  gBufferPoolTop = NULL;
  gBufferPoolNumFree = 0;

  pthread_mutex_init (&gBufferPoolMutex, NULL);
  pthread_cond_init (&gBufferPoolCondNotEmpty, NULL);

  for (i = 0; i<nNumBuffers; i++) {
    buf_element_t *pBuf;

    pBuf = xmalloc (sizeof (buf_element_t));

    if (pMultiBuffer != NULL) {
        pBuf->pMem = pMultiBuffer;
        pMultiBuffer += buf_size;
    }
    else
        pBuf->pMem = malloc_aligned (buf_size, alignment);

    pBuf->nMaxSize = buf_size;
    pBuf->next = NULL;
    
    buffer_pool_free (pBuf);
  }

  return &fifobuf_op;
}
