/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: demux_avi.c,v 1.10 2001/05/01 21:26:09 guenter Exp $
 *
 * demultiplexer for avi streams
 *
 * part of the code is taken from 
 * avilib (C) 1999 Rainer Johanni <Rainer@Johanni.de>
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>
#include <stdlib.h>

#include "xine.h"
#include "monitor.h"
#include "demux.h"
#include "utils.h"
#include "libw32dll/wine/mmreg.h"
#include "libw32dll/wine/avifmt.h"
#include "libw32dll/wine/vfw.h"

/* The following variable indicates the kind of error */

static uint32_t xine_debug;

uint32_t AVI_errno = 0;

typedef struct
{
  long pos;
  long len;
} video_index_entry;

typedef struct
{
  long pos;
  long len;
  long tot;
} audio_index_entry;

typedef struct
{
  long   width;             /* Width  of a video frame */
  long   height;            /* Height of a video frame */
  long   dwScale, dwRate;
  long   dwScale_audio, dwRate_audio;
  long   dwSampleSize;
  double fps;               /* Frames per second */
  
  char   compressor[8];     /* Type of compressor, 4 bytes + padding for 0 byte */
  long   video_strn;        /* Video stream number */
  long   video_frames;      /* Number of video frames */
  char   video_tag[4];      /* Tag of video data */
  long   video_posf;        /* Number of next frame to be read
			       (if index present) */
  long   video_posb;        /* Video position: byte within frame */
  
  long   a_fmt;             /* Audio format, see #defines below */
  long   a_chans;           /* Audio channels, 0 for no audio */
  long   a_rate;            /* Rate in Hz */
  long   a_bits;            /* bits per audio sample */
  long   audio_strn;        /* Audio stream number */
  long   audio_bytes;       /* Total number of bytes of audio data */
  long   audio_chunks;      /* Chunks of audio data in the file */
  char   audio_tag[4];      /* Tag of audio data */
  long   audio_posc;        /* Audio position: chunk */
  long   audio_posb;        /* Audio position: byte within chunk */
  
  long   pos;               /* position in file */
  long   n_idx;             /* number of index entries actually filled */
  long   max_idx;           /* number of index entries actually allocated */
  unsigned char (*idx)[16]; /* index entries (AVI idx1 tag) */
  video_index_entry   * video_index;
  audio_index_entry   * audio_index;
  BITMAPINFOHEADER      bih;
  char                  wavex[64];
  off_t  movi_start;
} avi_t;

typedef struct _demux_avi_globals {
  fifo_buffer_t       *mBufAudio;
  fifo_buffer_t       *mBufVideo;

  input_plugin_t      *mInput;

  avi_t               *avi;

  pthread_t            mThread;

  int                  mnAudioChannel;
  int                  mnStatus;

  uint32_t             video_step;
  uint32_t             nAvgBytesPerSec;
} demux_avi_globals_t ;

static demux_avi_globals_t gDemuxAvi;
static fifobuf_functions_t *Ffb;

#define AVI_ERR_SIZELIM      1     /* The write of the data would exceed
                                      the maximum size of the AVI file.
                                      This is more a warning than an error
                                      since the file may be closed safely */

#define AVI_ERR_OPEN         2     /* Error opening the AVI file - wrong path
                                      name or file nor readable/writable */

#define AVI_ERR_READ         3     /* Error reading from AVI File */

#define AVI_ERR_WRITE        4     /* Error writing to AVI File,
                                      disk full ??? */

#define AVI_ERR_WRITE_INDEX  5     /* Could not write index to AVI file
                                      during close, file may still be
                                      usable */

#define AVI_ERR_CLOSE        6     /* Could not write header to AVI file
                                      or not truncate the file during close,
                                      file is most probably corrupted */

#define AVI_ERR_NOT_PERM     7     /* Operation not permitted:
                                      trying to read from a file open
                                      for writing or vice versa */

#define AVI_ERR_NO_MEM       8     /* malloc failed */

#define AVI_ERR_NO_AVI       9     /* Not an AVI file */

#define AVI_ERR_NO_HDRL     10     /* AVI file has no has no header list,
                                      corrupted ??? */

#define AVI_ERR_NO_MOVI     11     /* AVI file has no has no MOVI list,
                                      corrupted ??? */

#define AVI_ERR_NO_VIDS     12     /* AVI file contains no video data */

#define AVI_ERR_NO_IDX      13     /* The file has been opened with
                                      getIndex==0, but an operation has been
                                      performed that needs an index */

static unsigned long str2ulong(unsigned char *str)
{
  return ( str[0] | (str[1]<<8) | (str[2]<<16) | (str[3]<<24) );
}

static unsigned long str2ushort(unsigned char *str)
{
  return ( str[0] | (str[1]<<8) );
}

static void long2str(unsigned char *dst, int n)
{
  dst[0] = (n    )&0xff;
  dst[1] = (n>> 8)&0xff;
  dst[2] = (n>>16)&0xff;
  dst[3] = (n>>24)&0xff;
}

static void AVI_close(avi_t *AVI)
{
  /* Even if there happened a error, we first clean up */

  if(AVI->idx) free(AVI->idx);
  if(AVI->video_index) free(AVI->video_index);
  if(AVI->audio_index) free(AVI->audio_index);
  free(AVI);
}

#define ERR_EXIT(x) \
{ \
   AVI_close(AVI); \
   AVI_errno = x; \
   return 0; \
}

#define PAD_EVEN(x) ( ((x)+1) & ~1 )

static int avi_sampsize(avi_t *AVI)
{
  int s;
  s = ((AVI->a_bits+7)/8)*AVI->a_chans;
  if(s==0) s=1; /* avoid possible zero divisions */
  return s;
}

static int avi_add_index_entry(avi_t *AVI, unsigned char *tag, 
			       long flags, long pos, long len)
{
  void *ptr;

  if(AVI->n_idx>=AVI->max_idx)
    {
      ptr = realloc((void *)AVI->idx,(AVI->max_idx+4096)*16);
      if(ptr == 0)
	{
	  AVI_errno = AVI_ERR_NO_MEM;
	  return -1;
	}
      AVI->max_idx += 4096;
      AVI->idx = (unsigned char((*)[16]) ) ptr;
    }

  /* Add index entry */

  memcpy(AVI->idx[AVI->n_idx],tag,4);
  long2str(AVI->idx[AVI->n_idx]+ 4,flags);
  long2str(AVI->idx[AVI->n_idx]+ 8,pos);
  long2str(AVI->idx[AVI->n_idx]+12,len);

  /* Update counter */

  AVI->n_idx++;

  return 0;
}

static avi_t *AVI_init(void) 
{
  avi_t *AVI;
  long i, n, idx_type;
  unsigned char *hdrl_data;
  long hdrl_len=0;
  long nvi, nai, ioff;
  long tot;
  int lasttag = 0;
  int vids_strh_seen = 0;
  int vids_strf_seen = 0;
  int auds_strh_seen = 0;
  int auds_strf_seen = 0;
  int num_stream = 0;
  char data[256];

  /* Create avi_t structure */

  AVI = (avi_t *) xmalloc(sizeof(avi_t));
  if(AVI==NULL)
    {
      AVI_errno = AVI_ERR_NO_MEM;
      return 0;
    }
  memset((void *)AVI,0,sizeof(avi_t));

  /* Read first 12 bytes and check that this is an AVI file */

  if( gDemuxAvi.mInput->read(data,12) != 12 ) ERR_EXIT(AVI_ERR_READ) ;
						
  if( strncasecmp(data  ,"RIFF",4) !=0 ||
      strncasecmp(data+8,"AVI ",4) !=0 ) ERR_EXIT(AVI_ERR_NO_AVI) ;
  /* Go through the AVI file and extract the header list,
     the start position of the 'movi' list and an optionally
     present idx1 tag */
  
  hdrl_data = 0;
  
  while(1) {
    if (gDemuxAvi.mInput->read(data,8) != 8 ) break; /* We assume it's EOF */
    
    n = str2ulong(data+4);
    n = PAD_EVEN(n);
    
    if(strncasecmp(data,"LIST",4) == 0)
      {
	if( gDemuxAvi.mInput->read(data,4) != 4 ) ERR_EXIT(AVI_ERR_READ)
						    n -= 4;
	if(strncasecmp(data,"hdrl",4) == 0)
	  {
	    hdrl_len = n;
	    hdrl_data = (unsigned char *) xmalloc(n);
	    if(hdrl_data==0) ERR_EXIT(AVI_ERR_NO_MEM)
			       if( gDemuxAvi.mInput->read(hdrl_data,n) != n ) ERR_EXIT(AVI_ERR_READ)
										}
	else if(strncasecmp(data,"movi",4) == 0)
	  {
	    AVI->movi_start = gDemuxAvi.mInput->seek(0,SEEK_CUR);
	    gDemuxAvi.mInput->seek(n,SEEK_CUR);
	  }
	else
	  gDemuxAvi.mInput->seek(n,SEEK_CUR);
      }
    else if(strncasecmp(data,"idx1",4) == 0)
      {
	/* n must be a multiple of 16, but the reading does not
	   break if this is not the case */
	
	AVI->n_idx = AVI->max_idx = n/16;
	AVI->idx = (unsigned  char((*)[16]) ) xmalloc(n);
	if(AVI->idx==0) ERR_EXIT(AVI_ERR_NO_MEM)
			  if( gDemuxAvi.mInput->read((char *)AVI->idx,n) != n ) ERR_EXIT(AVI_ERR_READ)
										  }
    else
      gDemuxAvi.mInput->seek(n,SEEK_CUR);
  }
  
  if(!hdrl_data) ERR_EXIT(AVI_ERR_NO_HDRL) ;
  if(!AVI->movi_start) ERR_EXIT(AVI_ERR_NO_MOVI) ;

  /* Interpret the header list */
  
  for(i=0;i<hdrl_len;)
    {
      /* List tags are completly ignored */
      
      if(strncasecmp(hdrl_data+i,"LIST",4)==0) { i+= 12; continue; }
      
      n = str2ulong(hdrl_data+i+4);
      n = PAD_EVEN(n);
      
      /* Interpret the tag and its args */
      
      if(strncasecmp(hdrl_data+i,"strh",4)==0)
	{
	  i += 8;
	  if(strncasecmp(hdrl_data+i,"vids",4) == 0 && !vids_strh_seen)
	    {
	      memcpy(AVI->compressor,hdrl_data+i+4,4);
	      AVI->compressor[4] = 0;
	      AVI->dwScale = str2ulong(hdrl_data+i+20);
	      AVI->dwRate  = str2ulong(hdrl_data+i+24);

	      if(AVI->dwScale!=0)
		AVI->fps = (double)AVI->dwRate/(double)AVI->dwScale;
	      gDemuxAvi.video_step = (long) (90000.0 / AVI->fps);
	      
	      AVI->video_frames    = str2ulong(hdrl_data+i+32);
	      AVI->video_strn = num_stream;
	      vids_strh_seen = 1;
	      lasttag = 1; /* vids */
	    }
	  else if (strncasecmp (hdrl_data+i,"auds",4) ==0 && ! auds_strh_seen)
	    {
	      AVI->audio_bytes = str2ulong(hdrl_data+i+32)*avi_sampsize(AVI);
	      AVI->audio_strn = num_stream;
	      AVI->dwScale_audio = str2ulong(hdrl_data+i+20);
	      AVI->dwRate_audio  = str2ulong(hdrl_data+i+24);
	      AVI->dwSampleSize  = str2ulong(hdrl_data+i+44);
	      auds_strh_seen = 1;
	      lasttag = 2; /* auds */
	    }
	  else
	    lasttag = 0;
	  num_stream++;
	}
      else if(strncasecmp(hdrl_data+i,"strf",4)==0)
	{
	  i += 8;
	  if(lasttag == 1) {
	    /* printf ("size : %d\n",sizeof(AVI->bih)); */
	    memcpy (&AVI->bih, hdrl_data+i, sizeof(AVI->bih));
	    /* stream_read(demuxer->stream,(char*) &avi_header.bih,MIN(size2,sizeof(avi_header.bih))); */
	    AVI->width  = str2ulong(hdrl_data+i+4);
	    AVI->height = str2ulong(hdrl_data+i+8);
	    
	    /*
	      printf ("size : %d x %d (%d x %d)\n", AVI->width, AVI->height, AVI->bih.biWidth, AVI->bih.biHeight);
	      printf("  biCompression %d='%.4s'\n", AVI->bih.biCompression, 
	      &AVI->bih.biCompression);
	    */
	    vids_strf_seen = 1;
	  }
	  else if(lasttag == 2)
	    {
	      memcpy (&AVI->wavex, hdrl_data+i, n);

	      AVI->a_fmt   = str2ushort(hdrl_data+i  );
	      AVI->a_chans = str2ushort(hdrl_data+i+2);
	      AVI->a_rate  = str2ulong (hdrl_data+i+4);
	      AVI->a_bits  = str2ushort(hdrl_data+i+14);
	      gDemuxAvi.nAvgBytesPerSec = str2ulong (hdrl_data+i+8);
	      auds_strf_seen = 1;
	    }
	  lasttag = 0;
	}
      else
	{
	  i += 8;
	  lasttag = 0;
	}
      
      i += n;
    }
  
  free(hdrl_data);
  
  if(!vids_strh_seen || !vids_strf_seen || AVI->video_frames==0) ERR_EXIT(AVI_ERR_NO_VIDS)

								   AVI->video_tag[0] = AVI->video_strn/10 + '0';
  AVI->video_tag[1] = AVI->video_strn%10 + '0';
  AVI->video_tag[2] = 'd';
  AVI->video_tag[3] = 'b';

  /* Audio tag is set to "99wb" if no audio present */
  if(!AVI->a_chans) AVI->audio_strn = 99;

  AVI->audio_tag[0] = AVI->audio_strn/10 + '0';
  AVI->audio_tag[1] = AVI->audio_strn%10 + '0';
  AVI->audio_tag[2] = 'w';
  AVI->audio_tag[3] = 'b';

  gDemuxAvi.mInput->seek(AVI->movi_start,SEEK_SET);

  /* if the file has an idx1, check if this is relative
     to the start of the file or to the start of the movi list */

  idx_type = 0;

  if(AVI->idx)
    {
      long pos, len;

      /* Search the first videoframe in the idx1 and look where
         it is in the file */

      for(i=0;i<AVI->n_idx;i++)
	if( strncasecmp(AVI->idx[i],AVI->video_tag,3)==0 ) break;
      if(i>=AVI->n_idx) ERR_EXIT(AVI_ERR_NO_VIDS)

			  pos = str2ulong(AVI->idx[i]+ 8);
      len = str2ulong(AVI->idx[i]+12);

      gDemuxAvi.mInput->seek(pos,SEEK_SET);
      if(gDemuxAvi.mInput->read(data,8)!=8) ERR_EXIT(AVI_ERR_READ) ;
      if( strncasecmp(data,AVI->idx[i],4)==0 && str2ulong(data+4)==len )
	{
	  idx_type = 1; /* Index from start of file */
	}
      else
	{
	  gDemuxAvi.mInput->seek(pos+AVI->movi_start-4,SEEK_SET);
	  if(gDemuxAvi.mInput->read(data,8)!=8) ERR_EXIT(AVI_ERR_READ) ;
	  if( strncasecmp(data,AVI->idx[i],4)==0 && str2ulong(data+4)==len )
	    {
	      idx_type = 2; /* Index from start of movi list */
	    }
	}
      /* idx_type remains 0 if neither of the two tests above succeeds */
    }
  
  if(idx_type == 0)
    {
      /* we must search through the file to get the index */

      gDemuxAvi.mInput->seek( AVI->movi_start, SEEK_SET);

      AVI->n_idx = 0;

      while(1)
	{
	  if( gDemuxAvi.mInput->read(data,8) != 8 ) break;
	  n = str2ulong(data+4);

	  /* The movi list may contain sub-lists, ignore them */

	  if(strncasecmp(data,"LIST",4)==0)
	    {
	      gDemuxAvi.mInput->seek(4,SEEK_CUR);
	      continue;
	    }

	  /* Check if we got a tag ##db, ##dc or ##wb */

	  if( ( (data[2]=='d' || data[2]=='D') &&
		(data[3]=='b' || data[3]=='B' || data[3]=='c' || data[3]=='C') )
	      || ( (data[2]=='w' || data[2]=='W') &&
		   (data[3]=='b' || data[3]=='B') ) )
	    {
	      avi_add_index_entry(AVI,data,0,gDemuxAvi.mInput->seek(0,SEEK_CUR)-8,n);
	    }

	  gDemuxAvi.mInput->seek(PAD_EVEN(n),SEEK_CUR);
	}
      idx_type = 1;
    }

  /* Now generate the video index and audio index arrays */

  nvi = 0;
  nai = 0;

  for(i=0;i<AVI->n_idx;i++)
    {
      if(strncasecmp(AVI->idx[i],AVI->video_tag,3) == 0) nvi++;
      if(strncasecmp(AVI->idx[i],AVI->audio_tag,4) == 0) nai++;
    }

  AVI->video_frames = nvi;
  AVI->audio_chunks = nai;

  if(AVI->video_frames==0) ERR_EXIT(AVI_ERR_NO_VIDS)
			     AVI->video_index = (video_index_entry *) xmalloc(nvi*sizeof(video_index_entry));
  if(AVI->video_index==0) ERR_EXIT(AVI_ERR_NO_MEM)
			    if(AVI->audio_chunks)
			      {
				AVI->audio_index = (audio_index_entry *) xmalloc(nai*sizeof(audio_index_entry));
				if(AVI->audio_index==0) ERR_EXIT(AVI_ERR_NO_MEM)
							  }

  nvi = 0;
  nai = 0;
  tot = 0;
  ioff = idx_type == 1 ? 8 : AVI->movi_start+4;

  for(i=0;i<AVI->n_idx;i++)
    {
      if(strncasecmp(AVI->idx[i],AVI->video_tag,3) == 0)
	{
	  AVI->video_index[nvi].pos = str2ulong(AVI->idx[i]+ 8)+ioff;
	  AVI->video_index[nvi].len = str2ulong(AVI->idx[i]+12);
	  nvi++;
	}
      if(strncasecmp(AVI->idx[i],AVI->audio_tag,4) == 0)
	{
	  AVI->audio_index[nai].pos = str2ulong(AVI->idx[i]+ 8)+ioff;
	  AVI->audio_index[nai].len = str2ulong(AVI->idx[i]+12);
	  AVI->audio_index[nai].tot = tot;
	  tot += AVI->audio_index[nai].len;
	  nai++;
	}
    }

  AVI->audio_bytes = tot;

  /* Reposition the file */

  gDemuxAvi.mInput->seek(AVI->movi_start,SEEK_SET);
  AVI->video_posf = 0;
  AVI->video_posb = 0;

  return AVI;
}

static long AVI_frame_size(avi_t *AVI, long frame)
{
  if(!AVI->video_index)         { AVI_errno = AVI_ERR_NO_IDX;   return -1; }

  if(frame < 0 || frame >= AVI->video_frames) return 0;
  return(AVI->video_index[frame].len);
}

static void AVI_seek_start(avi_t *AVI)
{
  AVI->video_posf = 0;
  AVI->video_posb = 0;
}

static int AVI_set_video_position(avi_t *AVI, long frame)
{
  if(!AVI->video_index)         { AVI_errno = AVI_ERR_NO_IDX;   return -1; }

  if (frame < 0 ) frame = 0;
  AVI->video_posf = frame;
  AVI->video_posb = 0;
  return 0;
}
      
static int AVI_set_audio_position(avi_t *AVI, long byte)
{
  long n0, n1, n;

  if(!AVI->audio_index)         { AVI_errno = AVI_ERR_NO_IDX;   return -1; }

  if(byte < 0) byte = 0;

  /* Binary search in the audio chunks */

  n0 = 0;
  n1 = AVI->audio_chunks;

  while(n0<n1-1)
    {
      n = (n0+n1)/2;
      if(AVI->audio_index[n].tot>byte)
	n1 = n;
      else
	n0 = n;
    }

  AVI->audio_posc = n0;
  AVI->audio_posb = byte - AVI->audio_index[n0].tot;

  return 0;
}

static long AVI_read_audio(avi_t *AVI, char *audbuf, 
			   long bytes, int *bFrameDone)
{
  long nr, pos, left, todo;

  if(!AVI->audio_index)         { AVI_errno = AVI_ERR_NO_IDX;   return -1; }

  nr = 0; /* total number of bytes read */

  /* printf ("avi audio package len: %d\n", AVI->audio_index[AVI->audio_posc].len); */


  while(bytes>0)
    {
      left = AVI->audio_index[AVI->audio_posc].len - AVI->audio_posb;
      if(left==0)
	{
	  AVI->audio_posc++;
	  AVI->audio_posb = 0;
	  if (nr>0) {
	    *bFrameDone = 1;
	    return nr;
	  }
	  left = AVI->audio_index[AVI->audio_posc].len - AVI->audio_posb;
	}
      if(bytes<left)
	todo = bytes;
      else
	todo = left;
      pos = AVI->audio_index[AVI->audio_posc].pos + AVI->audio_posb;
      /* printf ("demux_avi: read audio from %d\n", pos); */
      if (gDemuxAvi.mInput->seek (pos, SEEK_SET)<0)
	return -1;
      if (gDemuxAvi.mInput->read(audbuf+nr,todo) != todo)
	{
	  AVI_errno = AVI_ERR_READ;
	  *bFrameDone = 0;
	  return -1;
	}
      bytes -= todo;
      nr    += todo;
      AVI->audio_posb += todo;
    }

  left = AVI->audio_index[AVI->audio_posc].len - AVI->audio_posb;
  *bFrameDone = (left==0);

  return nr;
}

static long AVI_read_video(avi_t *AVI, char *vidbuf, 
			   long bytes, int *bFrameDone)
{
  long nr, pos, left, todo;

  if(!AVI->video_index)         { AVI_errno = AVI_ERR_NO_IDX;   return -1; }

  nr = 0; /* total number of bytes read */

  while(bytes>0)
    {
      left = AVI->video_index[AVI->video_posf].len - AVI->video_posb;
      if(left==0)
	{
	  AVI->video_posf++;
	  AVI->video_posb = 0;
	  if (nr>0) {
	    *bFrameDone = 1;
	    return nr;
	  }
	  left = AVI->video_index[AVI->video_posf].len - AVI->video_posb;
	}
      if(bytes<left)
	todo = bytes;
      else
	todo = left;
      pos = AVI->video_index[AVI->video_posf].pos + AVI->video_posb;
      /* printf ("demux_avi: read video from %d\n", pos); */
      if (gDemuxAvi.mInput->seek (pos, SEEK_SET)<0) 
	return -1;
      if (gDemuxAvi.mInput->read(vidbuf+nr,todo) != todo)
	{
	  AVI_errno = AVI_ERR_READ;
	  *bFrameDone = 0;
	  return -1;
	}
      bytes -= todo;
      nr    += todo;
      AVI->video_posb += todo;
    }

  left = AVI->video_index[AVI->video_posf].len - AVI->video_posb;
  *bFrameDone = (left==0);
	 
  return nr;
}

static int demux_avi_next (void) {

  buf_element_t *pBuf;

  if (gDemuxAvi.avi->video_frames <= gDemuxAvi.avi->video_posf)
    return 0;

  if (gDemuxAvi.avi->audio_chunks <= gDemuxAvi.avi->audio_posc)
    return 0;

  pBuf = Ffb->buffer_pool_alloc ();

  pBuf->pContent = pBuf->pMem;
  pBuf->nDTS  = 0 ; /* FIXME */

  if (gDemuxAvi.avi->audio_index[gDemuxAvi.avi->audio_posc].pos<
      gDemuxAvi.avi->video_index[gDemuxAvi.avi->video_posf].pos) {

    /* read audio */
    xprintf (VERBOSE|DEMUX|VAVI, "demux_avi: audio \n");

    if (gDemuxAvi.avi->dwSampleSize==0)
      pBuf->nPTS = gDemuxAvi.avi->audio_posc * (double) gDemuxAvi.avi->dwScale_audio / gDemuxAvi.avi->dwRate_audio * 90000.0;
    else 
      pBuf->nPTS = (gDemuxAvi.avi->audio_index[gDemuxAvi.avi->audio_posc].tot+gDemuxAvi.avi->audio_posb)/gDemuxAvi.avi->dwSampleSize * (double) gDemuxAvi.avi->dwScale_audio / gDemuxAvi.avi->dwRate_audio * 90000.0;

    pBuf->nSize     = AVI_read_audio (gDemuxAvi.avi, pBuf->pMem, 2048, &pBuf->bFrameEnd);


    if (pBuf->nSize<0) {
      Ffb->buffer_pool_free (pBuf);
      return 0;
    }

    /* printf ("demux_avi: audio package pts=%d\n",pBuf->nPTS); */
    
    pBuf->nInputPos = gDemuxAvi.mInput->seek (0, SEEK_CUR);

    switch (gDemuxAvi.avi->a_fmt) {
    case 0x01:
      pBuf->nType     = BUF_LINERPCM;
      break;
    case 0x2000:
      pBuf->nType     = BUF_AC3AUDIO;
      break;
    case 0x50:
    case 0x55:
      pBuf->nType     = BUF_MPEGAUDIO;
      break;
    case 0x161:
      pBuf->nType     = BUF_MSAUDIO;
      break;
    default:
      printf ("demux_avi: unknown audio type 0x%lx =>exit\n", gDemuxAvi.avi->a_fmt);
      gDemuxAvi.mnStatus = DEMUX_FINISHED;
      pBuf->nType     = BUF_MPEGAUDIO;
      break;
    }

    Ffb->fifo_buffer_put (gDemuxAvi.mBufAudio, pBuf);

  } else {
    /* read video */
    xprintf (VERBOSE|DEMUX|VAVI, "demux_avi: video \n");

    /* pBuf->nPTS = 0; */
    pBuf->nPTS = gDemuxAvi.avi->video_posf * (double) gDemuxAvi.avi->dwScale / gDemuxAvi.avi->dwRate * 90000.0;

    pBuf->nSize = AVI_read_video (gDemuxAvi.avi, pBuf->pMem, 2048, &pBuf->bFrameEnd);
    if (pBuf->nSize<0) {
      Ffb->buffer_pool_free (pBuf);
      return 0;
    }

    pBuf->nType = BUF_AVIVIDEO ; 

    /* printf ("demux_avi: video package pts=%d\n",pBuf->nPTS); */

    Ffb->fifo_buffer_put (gDemuxAvi.mBufVideo, pBuf);
  }

  xprintf (VERBOSE|DEMUX|VAVI, "size : %d\n",pBuf->nSize);

  return (pBuf->nSize>0);
}

static void *demux_avi_loop (void *dummy) {
  buf_element_t *pBuf;

  do {
    if (!demux_avi_next())
      gDemuxAvi.mnStatus = DEMUX_FINISHED;

  } while (gDemuxAvi.mnStatus == DEMUX_OK) ;

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_STREAMEND;
  Ffb->fifo_buffer_put (gDemuxAvi.mBufVideo, pBuf);
  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_STREAMEND;
  Ffb->fifo_buffer_put (gDemuxAvi.mBufAudio, pBuf);

  xprintf (VERBOSE|DEMUX, "avi demux loop finished (status: %d)\n",
	   gDemuxAvi.mnStatus);

  return NULL;
}

static void demux_avi_stop (void) {
  void *p;

  gDemuxAvi.mnStatus = DEMUX_FINISHED;
  
  Ffb->fifo_buffer_clear(gDemuxAvi.mBufVideo);
  Ffb->fifo_buffer_clear(gDemuxAvi.mBufAudio);

  pthread_join (gDemuxAvi.mThread, &p);
}

static int demux_avi_get_status (void) {
  return gDemuxAvi.mnStatus;
}

static void demux_avi_start (input_plugin_t *input_plugin,
			     fifo_buffer_t *bufVideo, 
			     fifo_buffer_t *bufAudio,
			     fifo_buffer_t *bufSPU,
			     off_t pos) 
{
  buf_element_t *pBuf;

  gDemuxAvi.mInput       = input_plugin;
  gDemuxAvi.mBufVideo    = bufVideo;
  gDemuxAvi.mBufAudio    = bufAudio;

  gDemuxAvi.mnStatus     = DEMUX_OK;

  gDemuxAvi.avi = AVI_init();

  if (!gDemuxAvi.avi) {
    printf ("demux_avi: init failed, avi_errno=%d .\n", AVI_errno);
    gDemuxAvi.mnStatus = DEMUX_FINISHED;
    return;
  }

  printf ("demux_avi: video format = %s, audio format = 0x%lx\n",
	  gDemuxAvi.avi->compressor, gDemuxAvi.avi->a_fmt);

  printf ("demux_avi: %d audio chunks, %d video frames\n",
	  gDemuxAvi.avi->audio_chunks, gDemuxAvi.avi->video_frames);

  AVI_seek_start (gDemuxAvi.avi);

  /*
   * seek
   */

  /* seek audio */
  while (gDemuxAvi.avi->audio_index[gDemuxAvi.avi->audio_posc].pos < pos) {
    gDemuxAvi.avi->audio_posc++;
    if (gDemuxAvi.avi->audio_posc>gDemuxAvi.avi->audio_chunks) {
      gDemuxAvi.mnStatus = DEMUX_FINISHED;
      return;
    }
  }

  /* seek video */

  
  while (gDemuxAvi.avi->video_index[gDemuxAvi.avi->video_posf].pos < pos) {
    gDemuxAvi.avi->video_posf++;
    if (gDemuxAvi.avi->video_posf>gDemuxAvi.avi->video_frames) {
      gDemuxAvi.mnStatus = DEMUX_FINISHED;
      return;
    }
  }
  

  /* 
   * send start buffers
   */

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_RESET;
  Ffb->fifo_buffer_put (gDemuxAvi.mBufVideo, pBuf);
  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_RESET;
  Ffb->fifo_buffer_put (gDemuxAvi.mBufAudio, pBuf);

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->pContent = pBuf->pMem;
  gDemuxAvi.avi->bih.biSize = gDemuxAvi.video_step; /* HACK */
  memcpy (pBuf->pContent, &gDemuxAvi.avi->bih, sizeof (gDemuxAvi.avi->bih));
  pBuf->nSize = sizeof (gDemuxAvi.avi->bih);
  pBuf->nType = BUF_AVIINIT; 
  Ffb->fifo_buffer_put (gDemuxAvi.mBufVideo, pBuf);

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->pContent = pBuf->pMem;
  memcpy (pBuf->pContent, &gDemuxAvi.avi->wavex, 
	  sizeof (gDemuxAvi.avi->wavex));
  pBuf->nSize = sizeof (gDemuxAvi.avi->wavex);
  pBuf->nType = BUF_AVIINIT; 
  Ffb->fifo_buffer_put (gDemuxAvi.mBufAudio, pBuf);

  pthread_create (&gDemuxAvi.mThread, NULL, demux_avi_loop, NULL) ;
}

static void demux_avi_select_audio_channel (int nChannel) {
  gDemuxAvi.mnAudioChannel = nChannel;
}

static int demux_avi_can_handle (char *MRL) {
  char *ending;
  
  /*
   * check ending
   */

  ending = strrchr(MRL, '.');
  xprintf(VERBOSE|DEMUX, "demux_avi_can_handle: ending %s of %s\n", ending, MRL);
  
  if(!ending)
    return 0;
  
  if(strcasecmp(ending, ".avi") 
     && (strcasecmp(ending, ".AVI"))) {
    return 0;
  }

  return 1;
}

static void demux_avi_block_select_spu_channel (int nChannel) {
}

static demux_functions_t demux_avi_functions = {
  NULL,
  NULL,
  demux_avi_can_handle,
  demux_avi_start,
  demux_avi_stop,
  demux_avi_get_status,
  demux_avi_select_audio_channel,
  demux_avi_block_select_spu_channel
};

demux_functions_t *init_demux_avi(fifobuf_functions_t *f, uint32_t xd) {

  Ffb = f;
  xine_debug = xd;
  return &demux_avi_functions;
}
