/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: demux_mpeg_block.c,v 1.8 2001/04/29 22:05:30 siggi Exp $
 *
 * demultiplexer for mpeg 1/2 program streams
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <string.h>

#include "xine.h"
#include "monitor.h"
#include "demux.h"

#include "gui/gui_main.h"

static uint32_t xine_debug;

typedef struct _demux_mpeg_block_globals {
  fifo_buffer_t       *mBufAudio;
  fifo_buffer_t       *mBufVideo;
  fifo_buffer_t       *mBufSPU;

  input_plugin_t      *mInput;

  pthread_t            mThread;

  int                  mnAudioChannel;
  int                  mnSPUChannel;
  int                  mnStatus;
  
  int                  mnBlocksize;
} demux_mpeg_block_globals_t ;

static demux_mpeg_block_globals_t gDemuxMpegBlock;
static fifobuf_functions_t *Ffb;


static void demux_mpeg_block_parse_pack () {

  buf_element_t *pBuf;
  unsigned char *p;
  int            bMpeg1=0;
  uint32_t       i;
  uint32_t       nHeaderLen;
  uint32_t       nPTS;
  uint32_t       nDTS;
  uint32_t       nPacketLen;
  uint32_t       nStreamID;

  pBuf = Ffb->buffer_pool_alloc ();

  i = (*gDemuxMpegBlock.mInput->read) (pBuf->pMem, gDemuxMpegBlock.mnBlocksize);

  if (i<gDemuxMpegBlock.mnBlocksize) {
    
    char *next_mrl;

    /*
     * check if seamless branching is possible
     */

    if ((next_mrl = gui_get_next_mrl () )) {
      printf ("demux_mpeg_block: checking if we can branch to %s\n", next_mrl);
      if (gDemuxMpegBlock.mInput->is_branch_possible (next_mrl)) {

	printf ("demux_mpeg_block: branching\n");

	gDemuxMpegBlock.mInput->open (next_mrl);
	gui_notify_demux_branched ();

	i = gDemuxMpegBlock.mInput->read (pBuf->pMem, gDemuxMpegBlock.mnBlocksize);
	
	if (i<gDemuxMpegBlock.mnBlocksize) {
	  gDemuxMpegBlock.mnStatus = DEMUX_FINISHED;
          Ffb->buffer_pool_free (pBuf);
	  return ;
	}

      } else {
	gDemuxMpegBlock.mnStatus = DEMUX_FINISHED;
        Ffb->buffer_pool_free (pBuf);
	return ;
      }
    } else {
      gDemuxMpegBlock.mnStatus = DEMUX_FINISHED;
      Ffb->buffer_pool_free (pBuf);
      return ;
    }
  }

  p = pBuf->pMem; /* len = gDemuxMpegBlock.mnBlocksize; */

  if (p[3] == 0xBA) { /* program stream pack header */

    int nStuffingBytes;

    xprintf (VERBOSE|DEMUX, "program stream pack header\n");

    bMpeg1 = (p[4] & 0x40) == 0;

    if (bMpeg1) {

      p   += 12;

    } else { /* mpeg2 */

      nStuffingBytes = p[0xD] & 0x07;

      xprintf (VERBOSE|DEMUX, "%d stuffing bytes\n",nStuffingBytes);

      p   += 14 + nStuffingBytes;
    }
  }


  if (p[3] == 0xbb) { /* program stream system header */
    
    int nHeaderLen;

    xprintf (VERBOSE|DEMUX, "program stream system header\n");

    nHeaderLen = (p[4] << 8) | p[5];

    p    += 6 + nHeaderLen;
  }

  /* we should now have a PES packet here */

  if (p[0] || p[1] || (p[2] != 1)) {
    fprintf (stderr, "demux error! %02x %02x %02x (should be 0x000001) \n",p[0],p[1],p[2]);
    Ffb->buffer_pool_free (pBuf);
    return ;
  }

  nPacketLen = p[4] << 8 | p[5];
  nStreamID  = p[3];

  xprintf (VERBOSE|DEMUX, "packet id = %02x len = %d\n",nStreamID, nPacketLen);

  if (bMpeg1) {

    if (nStreamID == 0xBF) {
      Ffb->buffer_pool_free (pBuf);
      return ;
    }

    p   += 6; /* nPacketLen -= 6; */

    while ((p[0] & 0x80) == 0x80) {
      p++; 
      nPacketLen--;
      /* printf ("stuffing\n");*/
    }

    if ((p[0] & 0xc0) == 0x40) {
      /* STD_buffer_scale, STD_buffer_size */
      p += 2;
      nPacketLen -=2;
    }

    nPTS = 0; 
    nDTS = 0;
    if ((p[0] & 0xf0) == 0x20) {
      nPTS  = (p[ 0] & 0x0E) << 29 ;
      nPTS |=  p[ 1]         << 22 ;
      nPTS |= (p[ 2] & 0xFE) << 14 ;
      nPTS |=  p[ 3]         <<  7 ;
      nPTS |= (p[ 4] & 0xFE) >>  1 ;
      p   += 5;
      nPacketLen -=5;
    } else if ((p[0] & 0xf0) == 0x30) {
      nPTS  = (p[ 0] & 0x0E) << 29 ;
      nPTS |=  p[ 1]         << 22 ;
      nPTS |= (p[ 2] & 0xFE) << 14 ;
      nPTS |=  p[ 3]         <<  7 ;
      nPTS |= (p[ 4] & 0xFE) >>  1 ;
      nDTS  = (p[ 5] & 0x0E) << 29 ;
      nDTS |=  p[ 6]         << 22 ;
      nDTS |= (p[ 7] & 0xFE) << 14 ;
      nDTS |=  p[ 8]         <<  7 ;
      nDTS |= (p[ 9] & 0xFE) >>  1 ;
      p   += 10;
      nPacketLen -= 10;
    } else {
      p++; 
      nPacketLen --;
    }

  } else { /* mpeg 2 */

    if (p[7] & 0x80) { /* PTS avail */
      
      nPTS  = (p[ 9] & 0x0E) << 29 ;
      nPTS |=  p[10]         << 22 ;
      nPTS |= (p[11] & 0xFE) << 14 ;
      nPTS |=  p[12]         <<  7 ;
      nPTS |= (p[13] & 0xFE) >>  1 ;
      
    } else
      nPTS = 0;
    
    if (p[7] & 0x40) { /* PTS avail */
      
      nDTS  = (p[14] & 0x0E) << 29 ;
      nDTS |=  p[15]         << 22 ;
      nDTS |= (p[16] & 0xFE) << 14 ;
      nDTS |=  p[17]         <<  7 ;
      nDTS |= (p[18] & 0xFE) >>  1 ;
      
    } else
      nDTS = 0;


    nHeaderLen = p[8];

    p    += nHeaderLen + 9;
    nPacketLen -= nHeaderLen + 3;
  }

  xprintf (VERBOSE|DEMUX, "stream_id=%x len=%d pts=%d dts=%d\n", nStreamID, nPacketLen, nPTS, nDTS);

  if (nStreamID == 0xbd) {

    int nTrack,nSPUID;
    
    nTrack = p[0] & 0x0F; /* hack : ac3 track */

    if((p[0] & 0xE0) == 0x20) {
      nSPUID = (p[0] & 0x1f);

      xprintf(VERBOSE|DEMUX, "SPU PES packet, id 0x%03x\n",p[0] & 0x1f);

      if((gDemuxMpegBlock.mnSPUChannel >= 0)
	 && (gDemuxMpegBlock.mBufSPU != NULL)
	 && (nSPUID == gDemuxMpegBlock.mnSPUChannel)) {
        pBuf->pContent  = p+1;
        pBuf->nSize     = nPacketLen-1;
        pBuf->nType     = BUF_SPU;
        pBuf->nPTS      = nPTS;
        pBuf->nDTS      = nDTS ;
        pBuf->nInputPos = gDemuxMpegBlock.mInput->seek (0, SEEK_CUR);
        
        Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufSPU, pBuf);    

        return;
      }
    }

    if (((p[0]&0xF0) == 0x80) && (nTrack == gDemuxMpegBlock.mnAudioChannel)) {

      xprintf (VERBOSE|DEMUX|AC3, "ac3 PES packet, track %02x\n",nTrack);
      /* printf ( "ac3 PES packet, track %02x\n",nTrack);  */

      pBuf->pContent  = p+4;
      pBuf->nSize     = nPacketLen-4;
      pBuf->nType     = BUF_AC3AUDIO;
      pBuf->nPTS      = nPTS;
      pBuf->nDTS      = nDTS ;
      pBuf->nInputPos = gDemuxMpegBlock.mInput->seek (0, SEEK_CUR);

      Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufAudio, pBuf);
      
      return ;
    } else if (((p[0]&0xf0) == 0xa0) || (nTrack == (gDemuxMpegBlock.mnAudioChannel-16))) {

      int pcm_offset;

      xprintf (VERBOSE|DEMUX,"LPCMacket, len : %d %02x\n",nPacketLen-4, p[0]);  

      /* printf ("PCM!!!!!!!!!!!!!!!!!!!!!!!\n"); */

      for( pcm_offset=0; ++pcm_offset < nPacketLen-1 ; ){
	if ( p[pcm_offset] == 0x01 && p[pcm_offset+1] == 0x80 ) { /* START */
	  pcm_offset += 2;
	  break;
	}
      }
  
      pBuf->pContent  = p+pcm_offset;
      pBuf->nSize     = nPacketLen-pcm_offset;
      pBuf->nType     = BUF_LINERPCM;
      pBuf->nPTS      = nPTS;
      pBuf->nDTS      = nDTS ;
      pBuf->nInputPos = gDemuxMpegBlock.mInput->seek (0, SEEK_CUR);

      Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufAudio, pBuf);
      
      return ;
    }

  } else if ((nStreamID >= 0xbc) && ((nStreamID & 0xf0) == 0xe0)) {

    xprintf (VERBOSE|DEMUX, "video %d\n", nStreamID);

    pBuf->pContent = p;
    pBuf->nSize    = nPacketLen;
    pBuf->nType    = BUF_MPEG2VIDEO;
    pBuf->nPTS     = nPTS;
    pBuf->nDTS     = nDTS;

    Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufVideo, pBuf);

    return ;

  }  else if ((nStreamID & 0xe0) == 0xc0) {
    int nTrack;

    nTrack = nStreamID & 0x1f;

    xprintf (VERBOSE|DEMUX|MPEG, "mpg audio #%d", nTrack);

    if ((bMpeg1 && (nTrack == gDemuxMpegBlock.mnAudioChannel))
	|| (!bMpeg1 && (nTrack == (gDemuxMpegBlock.mnAudioChannel-8)))) {
      
      pBuf->pContent  = p;
      pBuf->nSize     = nPacketLen;
      pBuf->nType     = BUF_MPEGAUDIO;
      pBuf->nPTS      = nPTS;
      pBuf->nDTS      = nDTS;
      pBuf->nInputPos = gDemuxMpegBlock.mInput->seek (0, SEEK_CUR);
      
      Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufAudio, pBuf);

      return ;
    }

  } else {
    xprintf (VERBOSE | DEMUX, "unknown packet, id = %x\n",nStreamID);
  }

  Ffb->buffer_pool_free (pBuf);

  return ;
  
}

static void *demux_mpeg_block_loop (void *dummy) {
  buf_element_t *pBuf;

  do {

    demux_mpeg_block_parse_pack();
    
  } while (gDemuxMpegBlock.mnStatus == DEMUX_OK) ;

  xprintf (VERBOSE|DEMUX, "demux loop finished (status: %d)\n",
	   gDemuxMpegBlock.mnStatus);

  gDemuxMpegBlock.mnStatus = DEMUX_FINISHED;

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_STREAMEND;
  Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufVideo, pBuf);
  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_STREAMEND;
  Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufAudio, pBuf);

  return NULL;
}

static void demux_mpeg_block_stop () {
  void *p;
  gDemuxMpegBlock.mnStatus = DEMUX_FINISHED;

  Ffb->fifo_buffer_clear(gDemuxMpegBlock.mBufVideo);
  Ffb->fifo_buffer_clear(gDemuxMpegBlock.mBufAudio);
  Ffb->fifo_buffer_clear(gDemuxMpegBlock.mBufSPU);

  pthread_join (gDemuxMpegBlock.mThread, &p);
}

static int demux_mpeg_block_get_status () {
  return gDemuxMpegBlock.mnStatus;
}

static void demux_mpeg_block_start (input_plugin_t *input_plugin,
				    fifo_buffer_t *bufVideo, 
				    fifo_buffer_t *bufAudio,
				    fifo_buffer_t *bufSPU,
				    off_t pos) 
{
  buf_element_t *pBuf;

  gDemuxMpegBlock.mInput       = input_plugin;

  gDemuxMpegBlock.mBufVideo    = bufVideo;
  gDemuxMpegBlock.mBufAudio    = bufAudio;
  gDemuxMpegBlock.mBufSPU      = bufSPU;

  gDemuxMpegBlock.mnStatus     = DEMUX_OK;

  if (input_plugin->get_blocksize ())
    gDemuxMpegBlock.mnBlocksize = input_plugin->get_blocksize ();

  pos /= (off_t) gDemuxMpegBlock.mnBlocksize;
  pos *= (off_t) gDemuxMpegBlock.mnBlocksize;

  xprintf (VERBOSE|DEMUX, "=>seek to %Ld\n",pos);

  gDemuxMpegBlock.mInput->seek (pos, SEEK_SET);

  /* 
   * send reset buffer
   */

  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_RESET;
  Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufVideo, pBuf);
  pBuf = Ffb->buffer_pool_alloc ();
  pBuf->nType    = BUF_RESET;
  Ffb->fifo_buffer_put (gDemuxMpegBlock.mBufAudio, pBuf);

  /*
   * now start demuxing
   */

  pthread_create (&gDemuxMpegBlock.mThread, NULL, demux_mpeg_block_loop, NULL) ;
}

static void demux_mpeg_block_select_audio_channel (int nChannel) {
  gDemuxMpegBlock.mnAudioChannel = nChannel;
  printf ("select audio channel : %d\n", nChannel);
}

static void demux_mpeg_block_select_spu_channel (int nChannel) {
  gDemuxMpegBlock.mnSPUChannel = nChannel;
}

static int demux_mpeg_block_can_handle (char *MRL) {
  char *media;
  char *ending;

  /*
   * check media type
   */

  media = strstr(MRL, "://");
  if(media) {
    if(!strncmp(MRL, "dvd", 3)
       || !strncmp(MRL, "fifo", 4)
       || (((!strncmp(MRL, "stdin", 5) || !strncmp(MRL, "fifo", 4))
	     && (!strncmp((media+3), "mpeg2", 5) ))) 
       ) {
      gDemuxMpegBlock.mnBlocksize = 2048;
      return 1;
    }
    if(!strncmp(MRL, "tcp", 3)) {
      gDemuxMpegBlock.mnBlocksize = 2324;
      return 1;
    }
    if(!strncmp(MRL, "vcd", 3)) {
      gDemuxMpegBlock.mnBlocksize = 2324;
      return 1;
    }
  } 
  else if(!strncmp(MRL, "-", 1)) {
    gDemuxMpegBlock.mnBlocksize = 2048;
    return 1;
  }
	  
  /*
   * check ending
   */

  ending = strrchr(MRL, '.');

  xprintf(VERBOSE|DEMUX, "demux_mpeg_nlock_can_handle: ending %s of %s\n",
	  ending ? ending :"(none)", MRL);
  
  if(!ending) {
    return 0;
  }

  if(!strcasecmp(ending, ".vob")) {
    gDemuxMpegBlock.mnBlocksize = 2048;
    return 1;
  }

  return 0;
}

static demux_functions_t demux_mpeg_block_functions = {
  NULL,
  NULL,
  demux_mpeg_block_can_handle,
  demux_mpeg_block_start,
  demux_mpeg_block_stop,
  demux_mpeg_block_get_status,
  demux_mpeg_block_select_audio_channel,
  demux_mpeg_block_select_spu_channel
};

demux_functions_t *init_demux_mpeg_block(fifobuf_functions_t *f, uint32_t xd) {
  gDemuxMpegBlock.mnSPUChannel = -1; /* Turn off SPU by default */

  Ffb = f;
  xine_debug = xd;
  return &demux_mpeg_block_functions;
}
