/*
 * decode.c
 *
 * Copyright (C) Aaron Holtzman <aholtzma@ess.engr.uvic.ca> - Nov 1999
 *
 * Decodes an MPEG-2 video stream.
 *
 * This file is part of mpeg2dec, a free MPEG-2 video stream decoder.
 *	
 * mpeg2dec is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 * 
 * mpeg2dec is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with GNU Make; see the file COPYING. If not, write to
 * the Free Software Foundation, 
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>

#include "xine.h"
#include "mpeg2.h"
#include "mpeg2_internal.h"
#include "attributes.h"
#include "monitor.h"

#ifdef __i386__
#include "mmx.h"
#include "sse.h"
#endif

extern uint32_t xine_debug;

//this is where we keep the state of the decoder
picture_t picture;

//global config struct
mpeg2_config_t config;

extern vo_driver_t *gVideoDriver;

//
//the current start code chunk we are working on
//we max out at 65536 bytes as that is the largest
//slice we will see in MP@ML streams.
// (we make no pretenses ofdecoding anything more than that)
static uint8_t chunk_buffer[224 * 1024 + 4];
static uint8_t *chunk_ptr = chunk_buffer;
static uint32_t shift = 0;

static uint32_t is_sequence_needed     = 1;

static int output_width                = 0;
static int output_height               = 0;
static int output_rate                 = 0;
static int output_ratio                = 0;

// sync stuff
static int nSkipFrames                 = 0;
static int nSkipBFrames                = 0;
static int bSkipSlices                 = 0;

static uint32_t given_video_pts;

int bMMX = 0;

int nVideoHours = 0;
int nVideoMins  = 0;
int nVideoSecs  = 0;

void mpeg2_init (void)
{

#ifdef __i386__
    int mm_supp;
#endif

    config.flags = 0;

#ifdef __i386__
#ifdef DEBUG
    mm_supp = 1; /* detection crashes with debug options => assume MMX */
#else
    mm_supp = mm_support ();
#endif
    xprintf (VERBOSE, "mm_support : %d\n", mm_supp);

    switch (mm_supp) {
    case 9:
      config.flags =  MM_ACCEL_X86_MMXEXT | MM_ACCEL_X86_MMX ;
      break;
    case 5:
      config.flags = MM_ACCEL_X86_3DNOW | MM_ACCEL_X86_MMX ;
      /* config.flags = MM_ACCEL_X86_MMX ; */
      break;
    case 3:
    case 1:
      config.flags = MM_ACCEL_X86_MMX ;
     break;
    }

    if (mm_supp>0) {
      bMMX = 1;
    }

#endif


    /* intialize the decoder state */
    shift = 0;
    is_sequence_needed = 1;

    header_state_init (&picture);
    idct_init ();
    motion_comp_init ();
    picture.forward_reference_frame_buf = NULL;
    picture.backward_reference_frame_buf = NULL;
    picture.throwaway_frame_buf = NULL;
    picture.current_frame_buf = NULL;
}

void decode_free_image_buffers () {

    if (picture.forward_reference_frame_buf) {
      vo_free_image_buffer (picture.forward_reference_frame_buf);
      picture.forward_reference_frame_buf = NULL;
    }

    if (picture.backward_reference_frame_buf) {
      vo_free_image_buffer (picture.backward_reference_frame_buf);
      picture.backward_reference_frame_buf = NULL;
    }

    if (picture.throwaway_frame_buf) {
      vo_free_image_buffer (picture.throwaway_frame_buf);
      picture.throwaway_frame_buf = NULL;
    }

    picture.current_frame_buf = NULL;

    picture.forward_reference_frame[0] = NULL;
    picture.forward_reference_frame[1] = NULL;
    picture.forward_reference_frame[2] = NULL;
    
    picture.backward_reference_frame[0]  = NULL;
    picture.backward_reference_frame[1]  = NULL;
    picture.backward_reference_frame[2]  = NULL;
    
    picture.current_frame[0] = NULL;
    picture.current_frame[1] = NULL;
    picture.current_frame[2] = NULL;
    
    picture.throwaway_frame[0]  = NULL;
    picture.throwaway_frame[1]  = NULL;
    picture.throwaway_frame[2]  = NULL;

}

void mpeg2_release_img_buffers () {

  decode_free_image_buffers ();
  
  bSkipSlices                 = 1;

}

void mpeg2_reset (vo_driver_t *video_out) {
  
    shift                       = 0;
    is_sequence_needed          = 0; /* FIXME */
    picture.picture_coding_type = 0;
    chunk_ptr                   = chunk_buffer;

    /*
     * free buffers
     */

    decode_free_image_buffers ();

    nSkipFrames                 = 0;
    nSkipBFrames                = 0;
    bSkipSlices                 = 1;

    /*
    header_state_init (&picture);
    idct_init ();
    motion_comp_init ();
    */

    xprintf (VERBOSE|MPEG, "mpeg2_reset done\n");
}

static void decode_reorder_frames ()
{
    if (picture.picture_coding_type != B_TYPE) {

        if (picture.forward_reference_frame_buf)
          vo_free_image_buffer (picture.forward_reference_frame_buf);

	/*
	 * make the backward reference frame the new forward reference frame
	 */

	picture.forward_reference_frame[0] =
	    picture.backward_reference_frame[0];
	picture.forward_reference_frame[1] =
	    picture.backward_reference_frame[1];
	picture.forward_reference_frame[2] =
	    picture.backward_reference_frame[2];
	picture.forward_reference_frame_buf = 
	    picture.backward_reference_frame_buf;

	/*
	 * allocate new backward reference frame
	 */


	picture.backward_reference_frame_buf = vo_alloc_image_buffer();
	picture.backward_reference_frame[0]  = picture.backward_reference_frame_buf->mem[0];
	picture.backward_reference_frame[1]  = picture.backward_reference_frame_buf->mem[1];
	picture.backward_reference_frame[2]  = picture.backward_reference_frame_buf->mem[2];
	picture.backward_reference_frame_buf->PTS       = 0;
	picture.backward_reference_frame_buf->bFrameBad = 1;

	/*
	 * make it the current frame
	 */

	picture.current_frame[0] = picture.backward_reference_frame[0];
	picture.current_frame[1] = picture.backward_reference_frame[1];
	picture.current_frame[2] = picture.backward_reference_frame[2];
	picture.current_frame_buf = picture.backward_reference_frame_buf;

    } else {

	/*
	 * allocate new throwaway frame
	 */


	picture.throwaway_frame_buf = vo_alloc_image_buffer();
	picture.throwaway_frame[0]  = picture.throwaway_frame_buf->mem[0];
	picture.throwaway_frame[1]  = picture.throwaway_frame_buf->mem[1];
	picture.throwaway_frame[2]  = picture.throwaway_frame_buf->mem[2];
	picture.throwaway_frame_buf->PTS       = 0;
	picture.throwaway_frame_buf->bFrameBad = 1;

	/*
	 * make it the current frame
	 */

	picture.current_frame[0] = picture.throwaway_frame[0];
	picture.current_frame[1] = picture.throwaway_frame[1];
	picture.current_frame[2] = picture.throwaway_frame[2];
	picture.current_frame_buf= picture.throwaway_frame_buf;

    }

}


static int parse_chunk (int code, uint8_t * buffer, uint32_t pts)
{
    int is_frame_done = 0;
    int bFlipPage = 0;

    /* printf ("mpeg2dec : parse_chunk, code=%02x\n",code); */

    if (is_sequence_needed && code != 0xb3)	/* b3 = sequence_header_code */
	return 0;

    /* stats_header (code, buffer); */


    switch (code) {
    case 0x00:	/* picture_start_code */

      bFlipPage = (picture.picture_structure == FRAME_PICTURE) || (picture.second_field) ;

      if (bFlipPage) {


        if (picture.picture_coding_type==B_TYPE) {
	  
	  if (picture.throwaway_frame_buf) {
	    vo_queue_frame(picture.throwaway_frame_buf);
	    vo_free_image_buffer (picture.throwaway_frame_buf);
	  }
	  picture.throwaway_frame_buf = NULL;
	  
        } else {
	  
	  if (picture.forward_reference_frame_buf) {
	    vo_queue_frame(picture.forward_reference_frame_buf);
	  }
	}
      }

      header_process_picture_header (&picture, buffer);
      /*
      printf ("###> coding type: %d\n", picture.picture_coding_type);
      */
      if (bFlipPage)
	decode_reorder_frames ();

      if (given_video_pts) {
	if (picture.current_frame_buf)
	  picture.current_frame_buf->PTS = given_video_pts;
	/*
	  printf ("pts for image of type %d given : %d\n",
	  picture.picture_coding_type, pts);
	*/
	given_video_pts = 0;
      }
      

      if ( (picture.picture_structure == FRAME_PICTURE) || (!picture.second_field)) {
	if (picture.current_frame_buf)	
	  picture.current_frame_buf->nType = picture.picture_coding_type;
      
	switch (picture.picture_coding_type) {
	case B_TYPE:

	  if (nSkipFrames || nSkipBFrames) {
	    bSkipSlices = 1;
	    if (nSkipFrames>0)
	      nSkipFrames--;
	    if (nSkipBFrames>0)
	      nSkipBFrames--;

	    return 0;
	  }
	  if (!picture.forward_reference_frame_buf
	      || !picture.backward_reference_frame_buf
	      || picture.forward_reference_frame_buf->bFrameBad 
	      || picture.backward_reference_frame_buf->bFrameBad) {
	    bSkipSlices = 1;

	    return 0;
	  }
	  bSkipSlices = 0;
	  break;
	case P_TYPE:

	  if (nSkipFrames) {
	    bSkipSlices = 1;
	    if (nSkipFrames>0)
	      nSkipFrames--;

	    return 0;
	  }
	  if (!picture.forward_reference_frame_buf
	      || picture.forward_reference_frame_buf->bFrameBad) {
	    bSkipSlices = 1;

	    return 0;
	  }
	  bSkipSlices = 0;
	  break;
	case I_TYPE:

	  if (nSkipFrames) {
	    if (nSkipFrames>0)
	      nSkipFrames--;
	    bSkipSlices = 1;

	    return 0;
	  }
	  bSkipSlices = 0;
	  break;
	}
      }
      break;

    case 0xb3:	/* sequence_header_code */
      header_process_sequence_header (&picture, buffer);
      is_sequence_needed = 0;

      if (( picture.coded_picture_width != output_width) 
	  || (picture.coded_picture_height != output_height)
	  || (picture.aspect_ratio_information != output_ratio)
	  || (picture.frame_rate_code != output_rate)) {
	  
	uint32_t video_step; /* in pts */

	decode_free_image_buffers ();

	switch (picture.frame_rate_code) {
	case 1: /* 23.976 fps */
	  video_step = 3913;
	  break;
	case 2: /* 24 fps */
	  video_step = 3750;
	  break;
	case 3: /* 25 fps */
	  video_step = 3600;
	  break;
	case 4: /* 29.97 fps */
	  video_step = 3003;
	  break;
	case 5: /* 30 fps */
	  video_step = 3000;
	  break;
	case 6: /* 50 fps */
	  video_step = 1800;
	  break;
	case 7: /* 59.94 fps */
	  video_step = 1525;
	  break;
	case 8: /* 60 fps */
	  video_step = 1509;
	  break;
	default:
	  fprintf (stderr, "invalid/unknown frame rate code : %d \n",
		   picture.frame_rate_code);
	  exit (1);
	}


	vo_set_image_format(picture.coded_picture_width,
			    picture.coded_picture_height,
			    picture.aspect_ratio_information,
			    IMGFMT_YV12, video_step);

	output_width  = picture.coded_picture_width;
	output_height = picture.coded_picture_height;
	output_rate   = picture.frame_rate_code;
	output_ratio  = picture.aspect_ratio_information;
      }
      break;

    case 0xb5:	/* extension_start_code */
      header_process_extension (&picture, buffer);
      break;

    case 0xb8:

      nVideoHours = ((buffer[0]&0x7c)>>2) - 1;
      nVideoMins  = ((buffer[0]&0x03)<<4)|((buffer[1]&0xF0)>>4);
      nVideoSecs  = ((buffer[1]&0x07)<<3)|((buffer[2]&0xe0)>>5);
      if( nVideoHours < 0 ) {  /* zero or countdown? */
         nVideoHours  = 0;
         nVideoMins   = 0;
         nVideoSecs   = 0;
      }
      /* 
      printf ("GOP %02d:%02d:%02d (%02d -- 0x%x)\n",
              (buffer[0]&0x7c)>>2,
              ((buffer[0]&0x03)<<4)|((buffer[1]&0xF0)>>4),
              ((buffer[1]&0x07)<<3)|((buffer[2]&0xe0)>>5),
              ((buffer[2]&0x1F)<<1)|((buffer[3]&0x80)>>7),
              *(uint32_t*)buffer );
      */
      break;
    default:
      if ((code < 0xb0) && code) {

	if (!bSkipSlices) { 

	  is_frame_done = slice_process (&picture, code, buffer);

	  /*
	  if (is_frame_done) 
	    printf ("frame done\n");
	    */
	  if (picture.current_frame_buf)
	    picture.current_frame_buf->bFrameBad = 0; 

	} else {
	  if (picture.current_frame_buf)
	    picture.current_frame_buf->bFrameBad = 1; 
	  /* printf ("Skip slices\n"); */
	}

      }
    }

    return is_frame_done;
}

int mpeg2_decode_data (uint8_t * current, uint8_t * end, uint32_t pts)
{
    static uint8_t code = 0xff;
    //static uint8_t chunk_buffer[65536];
    //static uint32_t shift = 0;

    uint8_t byte;
    int ret = 0;

    /*
    printf ("mpeg2_decode_data, pts = %d (%d %d %d %d ...)\n",pts,
    current[0], current[1], current[2], current[3]);
    */

    given_video_pts = pts;

    while (current != end) {
        while (1) {
            byte = *current++;
            if (shift == 0x00000100)
                break;
            *chunk_ptr++ = byte;
            shift = (shift | byte) << 8;

            if (current == end) {
#ifdef __i386__
                if (bMMX) 
                    emms();
#endif

                return ret;
            }
        }

        /* found start_code following chunk */

        ret += parse_chunk (code, chunk_buffer, pts);

        /* done with header or slice, prepare for next one */

        code = byte;
        chunk_ptr = chunk_buffer;
        shift = 0xffffff00;
    }

#ifdef __i386__
    if (bMMX) 
      emms();
#endif
    return ret;
}

int mpeg2_find_sequence_header (uint8_t * current, uint8_t * end)
{
    static uint8_t code = 0xff;
    //static uint8_t chunk_buffer[65536];
    //static uint8_t *chunk_ptr = chunk_buffer;
    //static uint32_t shift = 0;

    uint8_t byte;

    while (current != end) {
        while (1) {
            byte = *current++;
            if (shift == 0x00000100)
                break;
            *chunk_ptr++ = byte;
            shift = (shift | byte) << 8;

            if (current == end) {
	      return 0;
            }
        }

        /* found start_code following chunk */

	if (code == 0xb3) {	/* b3 = sequence_header_code */

	  header_process_sequence_header (&picture, chunk_buffer);
	  is_sequence_needed = 0;

	  if (( picture.coded_picture_width != output_width) 
	      || (picture.coded_picture_height != output_height)
	      || (picture.aspect_ratio_information != output_ratio)
	      || (picture.frame_rate_code != output_rate)) {
	  
	    uint32_t video_step; /* in pts */

	    decode_free_image_buffers ();

	    switch (picture.frame_rate_code) {
	    case 1: /* 23.976 fps */
	      video_step = 3913;
	      break;
	    case 2: /* 24 fps */
	      video_step = 3750;
	      break;
	    case 3: /* 25 fps */
	      video_step = 3600;
	      break;
	    case 4: /* 29.97 fps */
	      video_step = 3003;
	      break;
	    case 5: /* 30 fps */
	      video_step = 3000;
	      break;
	    case 6: /* 50 fps */
	      video_step = 1800;
	      break;
	    case 7: /* 59.94 fps */
	      video_step = 1525;
	      break;
	    case 8: /* 60 fps */
	      video_step = 1509;
	      break;
	    default:
	      fprintf (stderr, "invalid/unknown frame rate code : %d \n",
		       picture.frame_rate_code);
	      exit (1);
	    }


	    vo_set_image_format(picture.coded_picture_width,
				picture.coded_picture_height,
				picture.aspect_ratio_information,
				IMGFMT_YV12, video_step);

	    output_width  = picture.coded_picture_width;
	    output_height = picture.coded_picture_height;
	    output_rate   = picture.frame_rate_code;
	    output_ratio  = picture.aspect_ratio_information;
 
	    shift = 0;

	    chunk_ptr = chunk_buffer;

	    return 1;
	  }

	}
        /* done with chunk, prepare for next one */

        code = byte;
        chunk_ptr = chunk_buffer;
        shift = 0xffffff00;
    }

    return 0;
}

void mpeg2_set_skip_frames (int nFrames) {
  
  /*printf ("skip %d frames\n", nFrames);*/

  nSkipFrames = nFrames + nSkipBFrames;
  nSkipBFrames = 0;

}

void mpeg2_set_skip_bframes (int nFrames) {
  
  /*printf ("skip %d bframes\n", nFrames);*/

  nSkipBFrames = nFrames;
  
}

void mpeg2_set_skip_slices () {
  bSkipSlices  = 1;
}
