/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: main.c,v 1.58 2001/05/01 22:32:53 guenter Exp $
 *
 * xine main
 *
 */

/* required for getsubopt() */
#define _XOPEN_SOURCE 500

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include "demuxers/demux.h"
#include "xine.h"
#include "monitor.h"
#include "configfile.h"
#include "video_out/video_out.h"
#include "audio_out/audio_out.h"
#include "gui/gui_main.h"
#include "gui/gui_dnd.h"
#ifdef HAVE_ORBIT 
#include "corba/xine-server.h"
#endif
#ifdef HAVE_DXR3
#  include <libdxr3/api.h>
#endif
#ifdef HAVE_GETOPT_LONG
#  include <getopt.h>
#else
#  include "getopt.h"
#endif

uint32_t xine_debug;

extern Display        *gDisplay;
extern pthread_mutex_t gXLock;
extern vo_driver_t    *gVideoDriver;
extern Window          gVideoWin;

/* playlist */
int                    nPlayNum;
int                    ncur;
char                  *playlist[100];

/* options args */
static const char *short_options = "?hSrR"
#ifdef HAVE_XV
 "s"
#endif
#ifdef HAVE_LIRC
 "L"
#endif
#ifdef DEBUG
 "d:"
#endif
#ifdef HAVE_DXR3
 "t:o::"
#endif
#ifdef HAVE_RSUB
 "r:"
#endif
 "u:a:A:p::";
static struct option long_options[] = {
  {"help"           , no_argument      , 0, 'h' },
  {"spdif"          , no_argument      , 0, 'S' },
#ifdef HAVE_XV
  {"shared-memory"  , no_argument      , 0, 's' },
#endif
#ifdef HAVE_DXR3
  {"tv-mode"        , required_argument, 0, 't' },
  {"overlay-mode"   , optional_argument, 0, 'o' },
#endif
#ifdef HAVE_RSUB
  {"rsub-script"    , optional_argument, 0, 'r' },
#endif
  {"audio-channel"  , required_argument, 0, 'a' },
  {"audio-driver"   , required_argument, 0, 'A' },
  {"spu-channel"    , required_argument, 0, 'u' },
  {"auto-play"      , optional_argument, 0, 'p' },
  {"force-realtime" , no_argument,       0, 'r' },
  {"no-realtime"    , no_argument,       0, 'R' },
#ifdef HAVE_LIRC
  {"no-lirc"        , no_argument      , 0, 'L' },
#endif
#ifdef debug
  {"debug"          , required_argument, 0, 'd' },
#endif
  {0                , no_argument      , 0,  0  }
};

extern const char* ao_available[];
int ao_use_driver = AO_DRIVER_UNSET;
int use_realtime_sync = 1;
int force_realtime_sync = 0;
int curAudioTrack = 0;
int curSPUChannel = -1;
int no_lirc = 0;

#ifdef HAVE_DXR3
int tv_mode;
int overlay_mode;
int aspect_ratio;
#endif

#ifdef HAVE_RSUB
int enable_rsub;
char* xine_rsub_script;
#endif

/* ------------------------------------------------------------------ */
/*
 * Clean a string (remove spaces and '=' at the begin,
 * and '\n', '\r' and spaces at the end.
 */
char *atoa(char *str) {
  char *pbuf;

  pbuf = str;

  while(*pbuf != '\0') pbuf++;

  while(pbuf > str) {
    if(*pbuf == '\r' || *pbuf == '\n' || *pbuf == '"') pbuf = '\0';
    pbuf--;
  }

  while(*pbuf == '=' || *pbuf == '"') pbuf++;

  return pbuf;
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
void show_banner(void) {

  printf("This is %s - a free video player v%s\n(c) 2000, 2001 by G. Bartsch and the xine project team\n", 
	 PACKAGE, VERSION);
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
void show_usage (void) {

  printf("\n");
  printf("Usage: %s [OPTIONS]... [MRL]\n", PACKAGE);
  printf("\n");
  printf("OPTIONS are:\n");
#ifdef HAVE_XV
  printf("  -s, --shared-memory          Use X shared memory extension "
	 "for output.\n");
#endif
  printf("  -S, --spdif                  AC3 output via SPDIF Port.\n");
#ifdef HAVE_RSUB
  printf("  -r, --rsub-script <file>     rSub subtitle script to play.\n");
#endif
  printf("  -u, --spu-channel <#>        Select SPU (subtitle) channel '#'.\n");
  printf("  -a, --audio-channel <#>      Select audio channel '#'.\n");
  printf("  -A, --audio-driver <drv>     Select audio driver (oss"
#ifdef HAVE_ALSA
	 ", alsa"
#endif
#ifdef HAVE_ESD
	 ", esd"
#endif
	 " or null).\n");
  printf("  -r, --force-realtime         Force realtime audio sync.\n");
  printf("  -R, --no-realtime            Don't use realtime audio sync.\n");
  printf("  -p, --auto-play [opt]        Play on start. Can be followed by:\n");
  printf("                                 'f': start in fullscreen mode,\n");
  printf("                                 'h': hide control panel,\n");
  printf("                                 'q': quit when play is done.\n");
  printf("                                 'd': retrieve playlist from DVD.\n");
  printf("                                 'v': retrieve playlist from VCD.\n");
#ifdef HAVE_LIRC
  printf("  -L, --no-lirc                Turn off LIRC support.\n");
#endif
#ifdef DEBUG
  printf("  -d, --debug <flags>          Debug mode for <flags> ('help' for list).\n");
#endif
  printf("\n");
  printf("valid MRLs (media resource locator) are:\n");
  printf("  File:  'path/foo.vob'\n");
  printf("         '/path/foo.vob'\n");
  printf("         'file://path/foo.vob'\n");
  printf("         'fifo://[[mpeg1|mpeg2]:/]path/foo'\n");
  printf("         'stdin://[mpeg1|mpeg2]' or '-' (mpeg2 only)\n");
  printf("  DVD:   'dvd://VTS_01_2.VOB'\n");
  printf("  VCD:   'vcd://<track number>'\n");
  printf("\n");
}
/* ------------------------------------------------------------------------- */
/*
 *
 */
#ifdef DEBUG
int handle_debug_subopt(char *sopt) {
  int subopt;
  char *str = sopt;
  char *val = NULL;
  const char *debuglvl[] = {
    "verbose", "metronom", "audio", "demux", 
    "input", "video", "pts", "mpeg", "avi", 
    "ac3", "loop", "gui",
    NULL
  };
  
  while((subopt = getsubopt(&str, debuglvl, &val)) != -1) {
    
    switch(subopt) {
    default:
      /* If debug feature is already enabled, with 
       * XINE_DEBUG envvar, turn it off 
       */
      if((xine_debug & 0x8000>>(subopt + 1)))
	xine_debug &= ~0x8000>>(subopt + 1);
      else
	xine_debug |= 0x8000>>(subopt + 1);
      break;
    }
  }

  if(val) {
    int i = 0;
    fprintf(stderr, "Valid debug flags are:\n\t");
    while(debuglvl[i] != NULL) {
      fprintf(stderr,"%s, ", debuglvl[i]);
      i++;
    }
    fprintf(stderr, "\b\b.\n");
    return 0;
  }

  return 1;
}
#endif
/* ------------------------------------------------------------------------- */
/*
 * Handle audio output multi-args, return first audio plugin load successed.
 * Return AO_UNSET on fail. 
 */

int handle_audioout_subopt(char *sopt) {
  int subopt;
  char *str = sopt;
  char *val = NULL;
  int aod = 0;
 
  while((subopt = getsubopt(&str, ao_available, &val)) != -1) {
    
    aod = subopt;

    switch(subopt) {
    case AO_DRIVER_NULL:
      return AO_DRIVER_NULL;

    case AO_DRIVER_OSS:
      return AO_DRIVER_OSS;

#ifdef HAVE_ALSA      
    case AO_DRIVER_ALSA:
      return AO_DRIVER_ALSA; 
#endif

#ifdef HAVE_ESD
    case AO_DRIVER_ESD:
      return AO_DRIVER_ESD; 
#endif

#ifdef HAVE_DXR3
    case AO_DRIVER_DXR3:
      return AO_DRIVER_DXR3;
#endif

    }
  }
  
  return AO_DRIVER_UNSET;
}


/* ------------------------------------------------------------------------- */
/*
 *
 */
int main(int argc, char *argv[]) {
  int   option_index = 0;
  int   no_xv;
  int	ao_mode=AO_MODE_STEREO;
  int  c = '?';
  char *filename = NULL;
  char *display_name = ":0.0";

#ifdef HAVE_XV
  no_xv = 0;
#else
  no_xv = 1;
#endif

#ifdef HAVE_DXR3
  tv_mode = DXR3_TVMODE_NTSC;
  overlay_mode = DXR3_OVERLAY_MODE_OFF;
  aspect_ratio = DXR3_ASPECTRATIO_4_3;
#endif

#ifdef HAVE_RSUB
  enable_rsub = 0;
  xine_rsub_script = NULL;
#endif

  show_banner();

  xine_debug = 0;
#ifdef DEBUG
  /* If XINE_DEBUG envvar is set, parse it */
  if(getenv("XINE_DEBUG") != NULL) {
    if(!(handle_debug_subopt(atoa((getenv("XINE_DEBUG"))))))
      exit(1);
  }
  
  /*
   * profiling
   */

  profiler_init();
  profiler_set_label (0, "xine_loop     ");
  profiler_set_label (1, "ac3 decoding  ");
  profiler_set_label (2, "mpeg2 decoding");
  profiler_set_label (3, "mp3 decoding  ");
  profiler_set_label (4, "video output  ");
  profiler_set_label (5, "draw frame    ");
  
#endif

  /*
   * initialize CORBA server
   */
#ifdef HAVE_ORBIT
  xine_server_init (&argc, argv);
#endif
  /*
   * parse command line
   */
  opterr = 0;
  while((c = getopt_long(argc, argv, short_options, 
			 long_options, &option_index)) != EOF) {
    switch(c) {

    case 'S': /* Use SPDIF Port for AC3 Stream */
      ao_mode = AO_MODE_AC3;
      break;

#ifdef HAVE_XV
    case 's': /* Use Xshm insteat of Xv (if Xv is present) */
      no_xv = 1;
      break;
#endif
      
    case 'a': /* Select audio channel */
      sscanf(optarg, "%i", &curAudioTrack);
      /* track_code = track_table[audio_track]; */
      break;

#ifdef HAVE_DXR3
    case 't': /* DXR3 tv mode */
      if (!strcasecmp(optarg, "ntsc")) {
	tv_mode = DXR3_TVMODE_NTSC;
      }
      else if (!strcasecmp(optarg, "pal")) {
	tv_mode = DXR3_TVMODE_PAL;
      }
      else if (!strcasecmp(optarg, "pal60")) {
	tv_mode = DXR3_TVMODE_PAL60;
      }
      break;

    case 'o': /* DXR3 overlay mode */
      overlay_mode = DXR3_OVERLAY_MODE_RECTANGLE;
      break;
#endif

#ifdef HAVE_RSUB
    case 'r':
      enable_rsub = 1;
      xine_rsub_script = strdup(atoa(optarg));
      break;
#endif
      
    case 'u': /* Select SPU channel */
      sscanf(optarg, "%i", &curSPUChannel);
      break;

    case 'A': /* Select audio driver */
      if(optarg != NULL) {
	if((ao_use_driver = 
	    handle_audioout_subopt(atoa(optarg))) == AO_DRIVER_UNSET) {
	  int i;
	  fprintf(stderr, "Error: available audio output drivers are:");
	  for(i = AO_DRIVER_NULL; ao_available[i] != NULL; i++)
	    fprintf(stderr, " %s,", ao_available[i]);
	  fprintf(stderr, "\b.\n");
	  exit(1);
	}
      }
      break;

    case 'r': /* try realtime sybc even if cap test fails */
      force_realtime_sync = 1;
      break;

    case 'R': /* don't use audio driver's realtime capabilities */
      use_realtime_sync = 0;
      break;

    case 'p':/* Play [[in fullscreen][then quit]] on start */
      xine_user_pref |= PLAY_ON_START;
      if(optarg != NULL) {
	if(strrchr(optarg, 'f')) {
	  xine_user_pref |= FULL_ON_START;
	}
	if(strrchr(optarg, 'h')) {
	  xine_user_pref |= HIDEGUI_ON_START;
	}
	if(strrchr(optarg, 'q')) {
	  xine_user_pref |= QUIT_ON_STOP;
	}
	if(strrchr(optarg, 'd')) {
	  xine_user_pref |= PLAY_FROM_DVD;
	}
	if(strrchr(optarg, 'v')) {
	  xine_user_pref |= PLAY_FROM_VCD;
	}
      }
      break;

#ifdef HAVE_LIRC
    case 'L': /* Disable LIRC support */
      no_lirc = 1;
      break;
#endif

#ifdef DEBUG      
    case 'd': /* Select debug levels */
      if(optarg != NULL) {
	if(!(handle_debug_subopt(atoa(optarg))))
	  exit(1);
      }
      break;
#endif
      
    case 'h': /* Display usage */
    case '?':
      show_usage();
      exit(1);
      break;
      
    default:
      show_usage();
      fprintf (stderr, "invalid argument %d => exit\n",c);
      exit(1);
    }
  }

  /* Extract filename (if specified) */
  if (optind < argc) 
    filename = argv[optind];

  config_file_init ();

  /*
   * init X11
   */

  
  if (!XInitThreads ()) {
    printf ("\nXInitThreads failed - looks like you don't have a thread-safe xlib.\n");
    exit (1);
  }
  
  /*
  pthread_mutex_init (&gXLock, NULL);
  */

  if(getenv("DISPLAY"))
    display_name = getenv("DISPLAY");

  gDisplay = XOpenDisplay(display_name);

  if (gDisplay == NULL) {
    fprintf(stderr,"Can not open display\n");
    exit(1);
  }

  /*
   * drag and drop
   */
  
  //  gui_init_dnd();
  //  gui_dnd_set_callback (xDnd, gui_dndcallback);
  
  /*
   * xine init
   */

  xine_init (no_xv, gui_status_callback, ao_use_driver, ao_mode);
  xine_select_audio_channel (curAudioTrack);
  xine_select_spu_channel (curSPUChannel);

  /*
   * start CORBA server thread
   */
#ifdef HAVE_ORBIT
  xine_server_start ();
#endif

  /*
   * hand control over to gui
   */

  gui_start(argc-optind, &argv[optind]);

#ifdef HAVE_ORBIT
  xine_server_exit();
#endif

  return 0;
}		

