/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: video_out.c,v 1.40 2001/04/29 22:05:30 siggi Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>

#include "video_out.h"
#include "metronom.h"
#include "monitor.h"
#include "../libmpeg2/mpeg2.h"
#include "../gui/gui_main.h"
#include "../libspudec/spudec.h"

#define STABLE

#define NUM_BUFFERS     20

extern Window     gVideoWin;
extern Display   *gDisplay;
extern uint32_t   xine_debug;

typedef struct img_buf_fifo_s {
  vo_image_buffer_t *first;
  vo_image_buffer_t *last;
  int                num_buffers;

  pthread_mutex_t    mutex;
  pthread_cond_t     bNotEmpty;
} img_buf_fifo_t;

typedef struct vo_priv_s {

  img_buf_fifo_t    *free_img_buf_queue;
  img_buf_fifo_t    *display_img_buf_queue;

  int                video_step;
  int                pts_per_half_frame;
  int                format;

  /* statistics */

  int                num_frames_delivered;
  int                num_frames_skipped;
  int                num_frames_discarded;

  int                width, height;
  int                bInitialized;

  pthread_t          displayThread;
  int                bDisplayThreadRun;

  int                display_window_visible;

  /* overlay images */

  vo_image_buffer_t *overlay_image;
  vo_image_buffer_t *mask_image;
  int                bOverlayImage;

  /* SPU */

  spudec_geometry   *spu_geom;
} vo_priv_t;

vo_driver_t *gVideoDriver;
vo_priv_t    gVO;

img_buf_fifo_t *new_img_buf_queue () {
  img_buf_fifo_t *queue;

  queue = (img_buf_fifo_t *) malloc (sizeof (img_buf_fifo_t));
  if( queue ) {
    queue->first       = NULL;
    queue->last        = NULL;
    queue->num_buffers = 0;
    pthread_mutex_init (&queue->mutex, NULL);
    pthread_cond_init  (&queue->bNotEmpty, NULL);
  }
  return queue;
}

void append_to_img_buf_queue (img_buf_fifo_t *queue, vo_image_buffer_t *img) {

#if defined( DEBUG ) || !defined( STABLE )
  if( !queue || !img ) {
    printf("video_out : append_to_img_buf_queue() bad parms -- img=0x%lx,"
             " queue=0x%lx :" 
             " bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long) img, (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue ? queue->num_buffers : -1,
             queue ? (long)queue->first : -1, 
             queue ? (long)queue->last : -1 );
    return;
  }
#endif

  pthread_mutex_lock (&queue->mutex);

  img->next = NULL;

  if (!queue->first) {
    queue->first = img;
    queue->last  = img;
    queue->num_buffers = 0;
  } 
  else if (queue->last) {
    queue->last->next = img;
    queue->last  = img;
  }
#if defined(DEBUG) || !defined(STABLE)
  else {
    pthread_cond_signal (&queue->bNotEmpty);
    pthread_mutex_unlock (&queue->mutex);
    printf("video_out : append_to_img_buf_queue() queue corrupt -- queue=0x%lx :"
             " bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    vo_exit();
  }
#endif

  queue->num_buffers++;

  pthread_cond_signal (&queue->bNotEmpty);
  pthread_mutex_unlock (&queue->mutex);
}

vo_image_buffer_t *remove_from_img_buf_queue (img_buf_fifo_t *queue) {

  vo_image_buffer_t *img;

  pthread_mutex_lock (&queue->mutex);

  while (!queue->first) {
    pthread_cond_wait (&queue->bNotEmpty, &queue->mutex);
  }

  img = queue->first;

  if (img) {
    queue->first = img->next;
    img->next = NULL;
    if (!queue->first) {
      queue->last = NULL;
      queue->num_buffers = 0;
      pthread_cond_init  (&queue->bNotEmpty, NULL);
    }
#if defined(DEBUG) || !defined(STABLE)
    else if ( queue->num_buffers <= 0 ){
      printf("video_out : remove_from_img_buf_queue() invalid num" 
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
               (long)queue, 
              queue->bNotEmpty.__c_lock.__status, 
              queue->num_buffers, (long)queue->first, (long)queue->last );
    }
#endif
    else {
      queue->num_buffers--;
    }
  }
#if defined(DEBUG) || !defined(STABLE)
  else if ( queue->last || queue->num_buffers 
            || queue->bNotEmpty.__c_lock.__status ) { 
    printf("video_out : remove_from_img_buf_queue() queue struct corrupt"
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    queue->first = NULL;
    queue->last = NULL;
    queue->num_buffers = 0;
    pthread_cond_init  (&queue->bNotEmpty, NULL);
  }
  else {
    printf("video_out : remove_from_img_buf_queue() queue empty"
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    queue->first = NULL;
    queue->last = NULL;
    queue->num_buffers = 0;
    pthread_cond_init  (&queue->bNotEmpty, NULL);
  }
#endif
    
  pthread_mutex_unlock (&queue->mutex);

  return img;
}

vo_image_buffer_t *remove_from_img_buf_queue_noblock (img_buf_fifo_t *queue) {

  vo_image_buffer_t *img = NULL;

  pthread_mutex_lock (&queue->mutex);

  img = queue->first;

  if (img) {
    queue->first = img->next;
    img->next = NULL;
    if (!queue->first) {
      queue->last = NULL;
      queue->num_buffers = 0;
      pthread_cond_init(&queue->bNotEmpty, NULL);
    }
#if defined(DEBUG) || !defined(STABLE)
    else if ( queue->num_buffers <= 0 ){
      printf("video_out : remove_from_img_buf_queue_noblock() negative num" 
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
               (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    }
#endif
    else {
      queue->num_buffers--;
    }
  }
#if defined(DEBUG) || !defined(STABLE)
  else if ( queue->last || queue->num_buffers ) { 
    printf("video_out : remove_from_img_buf_queue_noblock() failed"
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    queue->last = NULL;
    queue->num_buffers = 0;
    pthread_cond_init(&queue->bNotEmpty, NULL);
  }
#endif

  pthread_mutex_unlock (&queue->mutex);

  return img;
}

vo_image_buffer_t *remove_from_img_buf_queue_have_mutex (img_buf_fifo_t *queue) {

  vo_image_buffer_t *img = NULL;

  /*
   * Caller must already hold mutex, call is non-blocking
   */

  img = queue->first;

  if (img) {
    queue->first = img->next;
    img->next = NULL;
    if (!queue->first) {
      queue->last = NULL;
      queue->num_buffers = 0;
      pthread_cond_init(&queue->bNotEmpty, NULL);
    }
#if defined(DEBUG) || !defined(STABLE)
    else if ( queue->num_buffers <= 0 ){
      printf("video_out : remove_from_img_buf_queue_have_mutex() negative num" 
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
               (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    }
#endif
    else {
      queue->num_buffers--;
    }
  }
#if defined(DEBUG) || !defined(STABLE)
  else if ( queue->last || queue->num_buffers ) { 
    printf("video_out : remove_from_img_buf_queue_have_mutex() failed"
             " -- queue=0x%lx : bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",
             (long)queue, 
             queue->bNotEmpty.__c_lock.__status, 
             queue->num_buffers, (long)queue->first, (long)queue->last );
    queue->last = NULL;
    queue->num_buffers = 0;
    pthread_cond_init(&queue->bNotEmpty, NULL);
  }
#endif

  return img;
}


/*
void print_queue (img_buf_fifo_t *queue) {
  
  printf ("first : %d last : %d num : %d\n",
	  queue->first, queue->last, queue->num_buffers);

}
*/

void free_img_buffers () {
  vo_image_buffer_t *img;

  /*
   * free all images
   */

  mpeg2_release_img_buffers ();
  gVideoDriver->reset ();

  while ((img = remove_from_img_buf_queue_noblock (gVO.display_img_buf_queue))) {
    gVideoDriver->dispose_image_buffer(img);
  }

  while ((img = remove_from_img_buf_queue_noblock (gVO.free_img_buf_queue))) {
    gVideoDriver->dispose_image_buffer(img);
  }

}

void vo_exit () {
  free_img_buffers ();
  gVideoDriver->exit ();
}

void reallocate_img_buffers () {

  int i;
  vo_image_buffer_t *img;
  
  free_img_buffers ();

  /*
  print_queue (gVO.display_img_buf_queue);
  print_queue (gVO.free_img_buf_queue);
  */
  /*
   * allocate new images
   */

  for (i=0; i<NUM_BUFFERS; i++) {
    img = gVideoDriver->alloc_image_buffer ();
    if( img )
      append_to_img_buf_queue (gVO.free_img_buf_queue, img);
  }

}

void vo_set_image_format (uint32_t width, uint32_t height, uint32_t ratio, 
			  int format, int video_step) {
  

  if (gVideoDriver->set_image_format (width, height, ratio, format)) 
    reallocate_img_buffers ();

  gVO.width              = width;
  gVO.height             = height;
  gVO.format             = format;
  gVO.video_step         = video_step;
  gVO.pts_per_half_frame = gVO.video_step/2;

  /* Create overlay images */
  if(gVO.overlay_image != NULL)
    gVideoDriver->dispose_image_buffer (gVO.overlay_image);
  if(gVO.mask_image != NULL)
    gVideoDriver->dispose_image_buffer (gVO.mask_image); 

  gVO.overlay_image = gVideoDriver->alloc_image_buffer ();
  gVO.mask_image = gVideoDriver->alloc_image_buffer ();

  gVO.bOverlayImage = spudec_set_images(gVO.overlay_image, gVO.mask_image,
                                        width, height, format);

  metronom_set_video_rate (gVO.video_step);
}

vo_image_buffer_t* vo_alloc_image_buffer() {
  vo_image_buffer_t *img;

  img = remove_from_img_buf_queue (gVO.free_img_buf_queue);
#if defined(DEBUG) || !defined(STABLE) 
  if (!img) {
    printf("video_out : vo_alloc_image_buffer() failed -- free_img_buf_queue=0x%lx :"
             " bNotEmpty=%ld, num=%d, first=0x%lx, last=0x%lx\n",  
             (long)gVO.free_img_buf_queue,
             gVO.free_img_buf_queue->bNotEmpty.__c_lock.__status, 
             gVO.free_img_buf_queue->num_buffers,
             (long)gVO.free_img_buf_queue->first,
             (long)gVO.free_img_buf_queue->last );
    vo_exit(); 
  }
#endif
  
  xprintf (VERBOSE|VIDEO, "video_out : alloc image buffer: 0x%lx\n", img);

  pthread_mutex_lock (&img->mutex);
  img->bDisplayLock = 0;
  img->bDecoderLock = 1;
  img->bDriverLock  = 0;
  pthread_mutex_unlock (&img->mutex);

  gVO.bInitialized = 1;

  return img;
}

/*
 * Called from decoder when its through with a buffer.
 * Buffer may still be in display or driver queues
 */
void vo_free_image_buffer(vo_image_buffer_t *img) {

  pthread_mutex_lock (&img->mutex);
  img->bDecoderLock = 0; 

#if defined(DEBUG) || !defined(STABLE)
  if ( img->bDisplayLock && img->bDriverLock ) {  /* not both! */
     printf("video_out: *** vo_free_image_buffer - img->bDisplayLock=%d,"
             " img->DriverLock=%d ***\n",
             img->bDisplayLock, img->bDriverLock ); 
  }
#endif
  if ( !img->bDisplayLock && !img->bDriverLock ) 
    append_to_img_buf_queue (gVO.free_img_buf_queue, img);

  pthread_mutex_unlock (&img->mutex);
}

/*
 * Called from driver, bDisplay should have been
 * cleared before handing off to the driver. But
 * decoder may not be done with the buffer.
 */
void vo_image_drawn (vo_image_buffer_t *img) {

  pthread_mutex_lock (&img->mutex);
  img->bDriverLock = 0;

#if defined(DEBUG) || !defined(STABLE)
  if ( img->bDisplayLock ) {
     printf("video_out: *** vo_image_drawn - img->bDisplayLock=%d,"
             " img->DecoderLock=%d ***\n",
             img->bDisplayLock, img->bDecoderLock ); 
  }
#endif

  if ( !img->bDecoderLock ) 
    append_to_img_buf_queue (gVO.free_img_buf_queue, img);

  pthread_mutex_unlock (&img->mutex);
}

void vo_queue_frame (vo_image_buffer_t* img) {

  int32_t  diff;
  uint32_t cur_vpts;
  uint32_t pic_vpts ;

  pic_vpts = metronom_got_video_frame (img->PTS);
  img->PTS = pic_vpts;
  gVO.num_frames_delivered++;

  xprintf (VERBOSE|VIDEO,"video_out: got image. vpts for picture is %d\n", pic_vpts);

  cur_vpts = metronom_get_current_time();

  diff = pic_vpts - cur_vpts;

  xprintf (VERBOSE|VIDEO,"video_out:: delivery diff : %d\n",diff);

  if (cur_vpts>0) {

    if (diff < gVO.video_step*5 ) {
      if (diff < gVO.video_step) {
	mpeg2_set_skip_frames (5); 
	
	xprintf (VERBOSE|VIDEO,"      ======> skip frame\n");
	
	if (diff < 0) { 
	  mpeg2_set_skip_frames (5 - diff / gVO.video_step); 
	  xprintf (VERBOSE|VIDEO, "               ======> rejected\n");
	  
	  gVO.num_frames_discarded++;
	  return; 
	} 
	
      } else {
	mpeg2_set_skip_bframes ((gVO.video_step*6 - diff) / gVO.video_step); 
	
	xprintf (VERBOSE|VIDEO,"      ======> skip bframe\n");
      }
    }
  } /* else: we are probably in precaching mode */

  if (!img->bFrameBad) {
    /*
     * put frame into FIFO-Buffer
     */

    xprintf (VERBOSE|VIDEO, "frame is ok => appending to display buffer\n");

    pthread_mutex_lock (&img->mutex);
    img->bDisplayLock = 1;
    append_to_img_buf_queue (gVO.display_img_buf_queue, img);
    pthread_mutex_unlock (&img->mutex);

  } else
    gVO.num_frames_skipped++;

  /*
   * performance measurement
   */
  
  if (gVO.num_frames_delivered>199) {
    printf ("%d frames delivered, %d frames skipped, %d frames discarded\n", 
	    gVO.num_frames_delivered, gVO.num_frames_skipped, gVO.num_frames_discarded);

    gVO.num_frames_delivered = 0;
    gVO.num_frames_discarded = 0;
    gVO.num_frames_skipped   = 0;
  }
}

/* 
 * display window is hideable
 */

void vo_toggle_display_window_visibility (void) {
  
  if (gVO.display_window_visible) {
    if (is_gui_panel_visible()) {
      gVO.display_window_visible = 0;
      XUnmapWindow (gDisplay, gVideoWin);
    }
  } else {
    gVO.display_window_visible = 1;
    XMapRaised(gDisplay, gVideoWin); 
  }
}

int is_display_window_visible (void) {
  return gVO.display_window_visible;
}


void vo_set_fullscreen (int bFullscreen) {
  gVideoDriver->set_fullscreen (bFullscreen);
  if(gVideoDriver->is_fullscreen())
    gVO.display_window_visible = 1;
}

int vo_is_fullscreen () {
  return gVideoDriver->is_fullscreen ();
}

void video_timer_handler (int hubba) {

  signal (SIGALRM, video_timer_handler);

}

void vo_set_timer (uint32_t video_step) {
  struct itimerval tval;

  tval.it_interval.tv_sec  = 0;
  tval.it_interval.tv_usec = video_step*100000/90000;
  tval.it_value.tv_sec     = 0;
  tval.it_value.tv_usec    = video_step*100000/90000;

  /* printf ("video_out: tval.it_interval.tv_usec = %d\n", tval.it_interval.tv_usec); */
  
  if (setitimer(ITIMER_REAL, &tval, NULL)) {
    perror ("setitimer failed :");
  }
}

void *video_out_loop (void *dummy) {

  uint32_t           cur_pts;
  int                diff, pts=0;
  vo_image_buffer_t *img;
  uint32_t           video_step = 0;

  xprintf (VERBOSE|VIDEO, 
            "video_out: video_out_loop() thread start - pid=%d run=%d\n",
            getpid(), gVO.bDisplayThreadRun ); 

  gVO.bDisplayThreadRun = 1;

  signal (SIGALRM, video_timer_handler);

  while (gVO.bDisplayThreadRun) {
    /*
     * Occasionally a soft hang would occur with this thread
     * stuck in pause() with a full queue of images. GUI pause-pause
     * would restart this thread and play would continue.
     * Suspect the interval timer had expired before getting
     * back around to pause (see bug in Linux setitimer() manpage?).
     * Moved the step check and itimer set to before the pause
     * to avoid the problem. Version to always set itimer is to 
     * avoid race between check and pause (conditioned for Linux).
     *
     * But, are these calls thread safe? Perhaps 
     * the pthread_signal functions should be used instead?
     * Or, at least pthread_sigmask?
     */ 
#if !defined(__linux__) 
    uint32_t video_step_new;
    video_step_new = metronom_get_video_rate ();
    if ( video_step_new != video_step) {
      video_step = video_step_new;
      vo_set_timer (video_step);
    }
#else
    video_step = metronom_get_video_rate ();
    vo_set_timer (video_step);
#endif

    pause ();

    if ( !gVO.display_img_buf_queue->first )
      continue;

    cur_pts = metronom_get_current_time ();
    
    xprintf (VERBOSE|VIDEO, "video_out : vo_video_out called at audio pts %d\n", cur_pts);
    
#ifdef DEBUG 
    /*
     * print image queue
     *
     *    Note: with multiple threads this could be messy without
     *    the lock.
     */
    img = gVO.display_img_buf_queue->first;
    
    while (img) {
      pts = img->PTS;
      xprintf (VERBOSE|VIDEO, 
                "video_out : display queue contains image with pts %d (cur pts %u)\n",
                pts,cur_pts);

#if defined(DEBUG) || !defined(STABLE)
      if( img == img->next ) {
        printf("video_out: *** image with pts %d points to itself:"
                  " img=0x%lx, img->next0x%lx,"
                  " queue=0x%lx: bNotEmpty=%ld, num_buffers %d,"
                  " first=0x%lx, last=0x%lx ***\n",
		pts, (long)img, (long)img->next, 
                (long)gVO.display_img_buf_queue,
                gVO.display_img_buf_queue->bNotEmpty.__c_lock.__status,
                gVO.display_img_buf_queue->num_buffers,
		(long)gVO.display_img_buf_queue->first,
                (long)gVO.display_img_buf_queue->last );
        vo_exit();
      }
#endif
      img = img->next;
    }
#endif
    
    /*
     *  With the multiple video-loop threads bug it was possible
     *  for one thread to overtake another in the expiration
     *  loop. This resulted in the queue removes
     *  extracting a different img then the one intended
     *  to be discarded or delivered to the spu loop. Also, 
     *  video_reset() could interfere. The more extensive
     *  consistency checks and explict locking were added 
     *  before the above bugs were found, I'm leaving them
     *  in for now.
     */
 
    img = gVO.display_img_buf_queue->first;
    
    if (!img) {
      continue;
    }

    /*
     * throw away expired frames
     */
    
    do {

      pts = img->PTS;
      diff = cur_pts - pts;
      
      if (diff > gVO.pts_per_half_frame) {
	/*
	  xprintf (VERBOSE|VIDEO, "video_out : throwing away image with pts %d because "
	  "it's too old (diff : %d > %d).\n",pts,diff,
	  gVO.pts_per_half_frame);
	*/
	printf ("video_out: throwing away image with pts %d because "
		"it's too old (diff : %d > %d).\n",pts,diff,
		gVO.pts_per_half_frame);

	gVO.num_frames_discarded++;

        pthread_mutex_lock ( &gVO.display_img_buf_queue->mutex );
        if ( img == gVO.display_img_buf_queue->first ) {
          img = remove_from_img_buf_queue_have_mutex (gVO.display_img_buf_queue);
          pthread_mutex_lock (&img->mutex);

          img->bDisplayLock = 0;

          if (!img->bDecoderLock) 
             append_to_img_buf_queue (gVO.free_img_buf_queue, img);

          pthread_mutex_unlock (&img->mutex);
        }
        else {
          printf("video_out: *** head of queue changed in expiration loop:"
                     " was=0x%lx, is=0x%lx ***\n",
                      (long)img, (long)gVO.display_img_buf_queue->first );
        }
        img = gVO.display_img_buf_queue->first;
        pthread_mutex_unlock ( &gVO.display_img_buf_queue->mutex );

        if ( !img )
          diff = -1;
      }
    } while (diff > gVO.pts_per_half_frame); 

    /*
     * time to display frame 0 ?
     */

    if (diff<0) {
      continue;
    }

    /*
     * remove frame from display queue and show it
     */
    
    pthread_mutex_lock ( &gVO.display_img_buf_queue->mutex );
    if ( img == gVO.display_img_buf_queue->first )
      img = remove_from_img_buf_queue_have_mutex (gVO.display_img_buf_queue);
    else {
      printf("video_out: *** head of queue changed after expiration loop:"
                      " was=0x%lx, is=0x%lx ***\n",
                      (long)img, (long)gVO.display_img_buf_queue->first );
      pthread_mutex_unlock ( &gVO.display_img_buf_queue->mutex );
      img = NULL; 
    }
    pthread_mutex_unlock ( &gVO.display_img_buf_queue->mutex );

    if (!img)
      continue;

    xprintf (VERBOSE|VIDEO, "video_out : displaying image with pts = %d (diff=%d)\n", pts, diff);
    
    pthread_mutex_lock (&img->mutex);

    if (!img->bDisplayLock)
      xprintf (VERBOSE|VIDEO, 
                "video_out: ALERT! frame was not locked for display queue\n");

    img->bDisplayLock = 0; 
    img->bDriverLock = 1;

    pthread_mutex_unlock (&img->mutex);

    /* 
     * Mix in SPU 
     */
     
    /* Tick SPUdec */
    spudec_tick();

    if(gVO.bOverlayImage) {
      /* This code is pretty nasty but quite quick which is important here! */
      /* APPROACH: Since 32bit processors hande data, well 32bits at a time,
       * we overlay the image word by word rather than byte by byte and then
       * tidy up the odd bytes at the end. This effectively cuts the number of
       * loop itterations by a factor of 4. */
      
      /* FIXME: Optimise for 64bit machines as well? */
      
      if((gVO.spu_geom->bIsVisible) && 
         (gVO.format == IMGFMT_YV12)) {
        /* Overlay the image. */
        long i, off;
        uint8_t  *img_l_8,  *ovl_l_8,  *msk_l_8;
        uint32_t *img_l_32, *ovl_l_32, *msk_l_32;
        uint8_t  *img_y_8,  *ovl_y_8,  *msk_y_8;
        uint32_t *img_y_32, *ovl_y_32, *msk_y_32;
        uint8_t  *img_v_8,  *ovl_v_8,  *msk_v_8;
        uint32_t *img_v_32, *ovl_v_32, *msk_v_32;

        img_l_8 = img->mem[0];
        ovl_l_8 = gVO.overlay_image->mem[0];
        msk_l_8 = gVO.mask_image->mem[0];
        img_l_32 = ((uint32_t*)img->mem[0]);
        ovl_l_32 = ((uint32_t*)gVO.overlay_image->mem[0]);
        msk_l_32 = ((uint32_t*)gVO.mask_image->mem[0]);
        img_y_8 = img->mem[1];
        ovl_y_8 = gVO.overlay_image->mem[1];
        msk_y_8 = gVO.mask_image->mem[1];
        img_y_32 = ((uint32_t*)img->mem[1]);
        ovl_y_32 = ((uint32_t*)gVO.overlay_image->mem[1]);
        msk_y_32 = ((uint32_t*)gVO.mask_image->mem[1]);
        img_v_8 = img->mem[2];
        ovl_v_8 = gVO.overlay_image->mem[2];
        msk_v_8 = gVO.mask_image->mem[2];
        img_v_32 = ((uint32_t*)img->mem[2]);
        ovl_v_32 = ((uint32_t*)gVO.overlay_image->mem[2]);
        msk_v_32 = ((uint32_t*)gVO.mask_image->mem[2]);

        /* luminance */
        for(i=gVO.width*gVO.spu_geom->start_row; 
            i<=gVO.width*gVO.spu_geom->end_row; i+=gVO.width) {
          /* i is address of begining of line. */

        /* Firstly, draw the start odd bytes. */
          for(off = i+gVO.spu_geom->start_col; (off & 3) != 0; off++) {
            if(msk_l_8[off] != 0xff) {
              img_l_8[off] &= msk_l_8[off];
              img_l_8[off] |= ovl_l_8[off];
            }
          }

          /* Now words */
          for(; off<=i+gVO.spu_geom->end_col-3; off+=4) {
            if(msk_l_32[off>>2] != 0xffffffff) {
              img_l_32[off>>2] &= msk_l_32[off>>2];
              img_l_32[off>>2] |= ovl_l_32[off>>2];
            }
          }
          off -= 4;

        /* Now end odd bytes */
          for(; off<=i+gVO.spu_geom->end_col; off++) {
            if(msk_l_8[off] != 0xff) {
              img_l_8[off] &= msk_l_8[off];
              img_l_8[off] |= ovl_l_8[off];
            }
          }
        }
        /* colour */
        for(i=(gVO.width>>1)*(gVO.spu_geom->start_row>>1); 
            i<=(gVO.width>>1)*(gVO.spu_geom->end_row>>1); i+=(gVO.width)>>1) {
          /* i is address of begining of line. */

        /* Firstly, draw the start odd bytes. */
          for(off = i+((gVO.spu_geom->start_col)>>1); (off & 3) != 0; off++) {
            if(msk_y_8[off] != 0xff) {
              img_y_8[off] &= msk_y_8[off];
              img_y_8[off] |= ovl_y_8[off];
            }
            if(msk_v_8[off] != 0xff) {
              img_v_8[off] &= msk_v_8[off];
              img_v_8[off] |= ovl_v_8[off];
            }
          }

          /* Now words */
          for(; off<=i+((gVO.spu_geom->end_col)>>1)-3; off+=4) {
            if(msk_y_32[off>>2] != 0xffffffff) {
              img_y_32[off>>2] &= msk_y_32[off>>2];
              img_y_32[off>>2] |= ovl_y_32[off>>2];
            }
            if(msk_v_32[off>>2] != 0xffffffff) {
              img_v_32[off>>2] &= msk_v_32[off>>2];
              img_v_32[off>>2] |= ovl_v_32[off>>2];
            }
          }
          off -= 4;

        /* Final end odd bytes */
          for(; off<=i+((gVO.spu_geom->end_col)>>1); off++) {
            if(msk_y_8[off] != 0xff) {
              img_y_8[off] &= msk_y_8[off];
              img_y_8[off] |= ovl_y_8[off];
            }
            if(msk_v_8[off] != 0xff) {
              img_v_8[off] &= msk_v_8[off];
              img_v_8[off] |= ovl_v_8[off];
            }
          }
        }
      }
    }

    xprintf (VERBOSE|VIDEO,
              "video_out : passing to video driver, image with pts = %d\n", pts);

    gVideoDriver->display_frame (img);
    
    /* 
     * video driver will free the image when its done with it 
     */
  }

  xprintf (VERBOSE|VIDEO,
            "video_out: video_out_loop() thread exiting - pid=%d run=%d\n",
            getpid(), gVO.bDisplayThreadRun ); 
  return NULL;
}

void vo_handle_event (XEvent *event) {
  gVideoDriver->handle_event (event);
}

void vo_set_logo_mode (int bLogoMode) {

  xprintf (VERBOSE|VIDEO, 
             "video_out: set_logo_mode() %d - run %d \n",
             bLogoMode,gVO.bDisplayThreadRun);

  gVideoDriver->set_logo_mode (bLogoMode);

  if (bLogoMode) {
    void *p;

    if ( gVO.displayThread || gVO.bDisplayThreadRun) {
      gVO.bDisplayThreadRun = 0;
      xprintf (VERBOSE|VIDEO, 
                  "video_out: set_logo_mode() joining thread %ld\n",
                  gVO.displayThread);
      pthread_join (gVO.displayThread, &p);
      gVO.displayThread = 0;
    }

  }
  else {
    if ( gVO.displayThread || gVO.bDisplayThreadRun ) {
      void *p;
      xprintf (VERBOSE|VIDEO, 
                "video_out: set_logo_mode() thread exists %ld\n",
                gVO.displayThread);
      gVO.bDisplayThreadRun = 0;
      xprintf (VERBOSE|VIDEO,
                "video_out: set_logo_mode() joining thread %ld\n",
                gVO.displayThread);
      pthread_join (gVO.displayThread, &p);
      gVO.displayThread = 0;
    }
    pthread_create (&gVO.displayThread, NULL, video_out_loop, NULL) ;
    xprintf (VERBOSE|VIDEO,
               "video_out: set_logo_mode() created thread %ld\n",
               gVO.displayThread);
  }
}

void vo_reset (void) {
  vo_image_buffer_t *img;
  void *p;

  xprintf (VERBOSE|VIDEO, "video_out: vo_reset() \n");

  if( gVO.bDisplayThreadRun || gVO.displayThread ) {
    gVO.bDisplayThreadRun = 0;  /* stop the thread */
    /* raise (SIGALRM); */
    xprintf (VERBOSE|VIDEO,
              "video_out: vo_reset() joining video_out_loop() thread %ld\n",
              gVO.displayThread);
    pthread_join (gVO.displayThread, &p);
    xprintf (VERBOSE|VIDEO,
              "video_out: vo_reset() video_out_loop() thread %ld joined\n",
              gVO.displayThread);
    gVO.displayThread        = 0;
  }

  while (
     (img = remove_from_img_buf_queue_noblock (gVO.display_img_buf_queue))
  ) {

#if defined(DEBUG) || !defined(STABLE)
    if (!img->bDisplayLock) {
      printf ("video_out: ALERT! frame was not locked in display queue\n");
    }
#endif

    pthread_mutex_lock ( &img->mutex );
    img->bDisplayLock = 0;

    if (!img->bDecoderLock) {
      append_to_img_buf_queue (gVO.free_img_buf_queue, img);
    }       
    pthread_mutex_unlock ( &img->mutex );
  }

  gVO.num_frames_delivered = 0;
  gVO.num_frames_skipped   = 0;
  gVO.num_frames_discarded = 0;
  gVO.bInitialized         = 0;   
}

void vo_display_cursor (int state) {

  gVideoDriver->hideshow_cursor(state);
}

void vo_set_aspect (int ratio) {
  free_img_buffers ();
  gVideoDriver->set_aspect (ratio);
  reallocate_img_buffers ();
}

int vo_get_aspect () {
  return gVideoDriver->get_aspect ();
}

void vo_init (int no_xv) {

  gVideoDriver = NULL;

  gVO.width                  = 720;
  gVO.height                 = 576;
  gVO.num_frames_delivered   = 0;
  gVO.num_frames_skipped     = 0;
  gVO.num_frames_discarded   = 0;
  gVO.free_img_buf_queue     = new_img_buf_queue ();
  gVO.display_img_buf_queue  = new_img_buf_queue ();
  gVO.bInitialized           = 0;   
  gVO.bOverlayImage          = 0;
  gVO.overlay_image          = NULL;
  gVO.mask_image             = NULL;
  gVO.spu_geom               = spudec_get_geometry();
  gVO.pts_per_half_frame     = 3000;   
  gVO.bDisplayThreadRun      = 0;
  gVO.displayThread          = 0;
  gVO.display_window_visible = 1;


  gVideoDriver = init_video_out_mga ();
  if (gVideoDriver) {
    printf ("Using Teletux Matrox G400 video extension for video output.\n");

    vo_set_image_format (720,576,1,IMGFMT_YV12,3000);

    return ;
  }

#ifdef HAVE_XV
  if (!no_xv) {
    gVideoDriver = init_video_out_xv ();
    if (gVideoDriver) {
      printf ("Using X Window System video extension for video output.\n");

      vo_set_image_format (720,405,3,IMGFMT_YV12,3000);

      return ;
    }
  }
#endif
  
  gVideoDriver = init_video_out_xshm ();
  if (gVideoDriver) {
    printf ("Using X Window System shared memory extension for video output.\n");

    vo_set_image_format (720,576,1,IMGFMT_YV12,3000);

    return ;
  }
  

  /*
  vo = init_video_out_x11 ();
  if (vo) {
    printf ("Using plain X11 for video output.\n");
    return vo;
  }
  */

  fprintf (stderr, "\nAll video output plugins failed. Exit.\n\n");
  exit (1);
}

void vo_hide_spu(void) {
  spudec_reset();
}
