/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: video_out.h,v 1.29 2001/04/29 22:05:30 siggi Exp $
 *
 *
 * xine version of video_out.h 
 * handles frame-ordering itself
 *
 * due to performance reasons, there is no wrapper function for
 *
 * notify_frame_area_ready 
 *
 * call this directly using gVideoDriver->notify_frame_area_ready
 *
 * for all other functions, use the wrappers below
 *
 */

#ifndef HAVE_VIDEO_OUT_H

#define HAVE_VIDEO_OUT_H

#include "xine.h"
#include "inttypes.h"
#include <X11/Xlib.h>
#include <pthread.h>

typedef struct vo_image_buffer_s /* public part, video drivers may add private fields */
{
  struct vo_image_buffer_s  *next;

  uint32_t                   PTS;
  int                        bFrameBad; /* e.g. frame skipped or based on skipped frame */
  uint8_t                   *mem[3];
  int                        nType;     /* I, B or P frame */

  int                        bDisplayLock, bDecoderLock, bDriverLock;
  pthread_mutex_t            mutex; /* so the various locks will be serialized */

  int                        nID; /* debugging purposes only */
} vo_image_buffer_t;


int is_display_window_visible(void);


/*
 * init xine video output, check for usable output drivers
 *
 * result in global variable vo_driver_t *gVideoDriver;
 *
 */

void vo_init (int no_xv) ;

/*
 * set video image format
 * display driver can allocate image buffers now
 *
 *    params : width      == width of video to display.
 *             height     == height of video to display.
 *             ratio      == aspect ration information
 *             format     == FOURCC descriptor of image format
 *             video_step == frame duration in 1/90000 sec
 */

void vo_set_image_format (uint32_t width, uint32_t height, uint32_t ratio, 
			  int format, int video_step);

#define IMGFMT_YV12 0x32315659
#define IMGFMT_YUY2 (('2'<<24)|('Y'<<16)|('U'<<8)|'Y')
#define IMGFMT_RGB  (('R'<<24)|('G'<<16)|('B'<<8))

/*
 * allocate an image buffer from display driver
 */
vo_image_buffer_t* vo_alloc_image_buffer();

/*
 * image buffer is no longer used by decoder
 */

void vo_free_image_buffer(vo_image_buffer_t *vo_img);

/*
 * output driver doesn't use this image any more
 */

void vo_image_drawn(vo_image_buffer_t *vo_img);

/*
 * notify driver that a certain area of the image is ready for copying
 *
 * return value: 0 => driver ignored this, 1 => driver copied section
 */

int vo_notify_frame_area_ready (vo_image_buffer_t *img, int nLine, int x0, int x1);

/*
 * add frame to display queue
 */

void vo_queue_frame (vo_image_buffer_t* img);


/*
 * toggle visibility of the video window
 */

void vo_toggle_display_window_visibility ();


/*
 * set display driver to fullscreen mode
 */

void vo_set_fullscreen (int bFullscreen);

/*
 * returns 1 if driver is in fullscreen mode
 */

int vo_is_fullscreen (void);


/*
 * handle events for video window (e.g. expose)
 */

void vo_handle_event (XEvent *event) ;

/*
 * turn logo mode on/off
 * when logo mode is off, the queue is checked regularly for frames to display
 */

void vo_set_logo_mode (int bLogoMode);

/*
 * reset output device (reset buffers etc.)
 */

void vo_reset (void);

/*
 * release shared memory
 */

void vo_exit (void);

/*
 * hide/show mouse cursor.
 */
void vo_display_cursor (int state);

/* get/set aspect ratio */

void vo_set_aspect (int ratio);
int vo_get_aspect ();

/*
 * Hide spu
 */
void vo_hide_spu(void);


/*
 * these are the functions every display driver has to implement:
 *
 * only notify_frame_portion_ready should be called directly.
 */

typedef struct vo_driver_s
{

  int (*get_capabilities) (void); /* see below */

  /* return 1 if image buffers need to be re-allocated */
  int (*set_image_format)(uint32_t width, uint32_t height, uint32_t ratio,
			  int format);

  vo_image_buffer_t* (*alloc_image_buffer) (void);

  void (*dispose_image_buffer) (vo_image_buffer_t *img);

  void (*process_macroblock) (vo_image_buffer_t *img, 
			      uint8_t *py, uint8_t *pu, uint8_t *pv, /* source */
			      int slice_offset, /* pos in dest frame */
			      int offset);

  void (*display_frame) (vo_image_buffer_t* img);

  void (*set_fullscreen) (int bFullscreen);

  int (*is_fullscreen) (void);

  void (*handle_event) (XEvent *event) ;

  void (*set_logo_mode) (int bLogoMode);

  void (*reset) (void);

  void (*hideshow_cursor) (int state);

  void (*set_aspect) (int ratio);

  int (*get_aspect) (void);

  void (*exit) (void);

  int (*get_hue_min) (void);
  int (*get_current_hue) (void);
  int (*set_hue) (int);
  int (*get_hue_max) (void);

  int (*get_saturation_min) (void);
  int (*get_current_saturation) (void);
  int (*set_saturation) (int);
  int (*get_saturation_max) (void);

  int (*get_brightness_min) (void);
  int (*get_current_brightness) (void);
  int (*set_brightness) (int);
  int (*get_brightness_max) (void);

  int (*get_contrast_min) (void);
  int (*get_current_contrast) (void);
  int (*set_contrast) (int);
  int (*get_contrast_max) (void);

  int (*get_colorkey_min) (void);
  int (*get_current_colorkey) (void);
  int (*set_colorkey) (int);

  void (*reset_settings) (void);
  void (*save_settings) (void);

} vo_driver_t;

/* capabilities */

/* driver copies image (i.e. converts it to 
   rgb buffers in the private fields of image buffer) */
#define VO_CAP_COPIES_IMAGE 0x00000001
#define VO_CAP_RGB          0x00000002 /* driver can handle 24bit rgb pictures */
#define VO_CAP_YV12         0x00000004 /* driver can handle YUV 4:2:0 pictures */
#define VO_CAP_YUY2         0x00000008 /* driver can handle YUY2      pictures */
#define VO_CAP_HUE          0x00000010 /* driver can set HUE value             */
#define VO_CAP_SATURATION   0x00000020 /* driver can set SATURATION value      */
#define VO_CAP_BRIGHTNESS   0x00000040 /* driver can set BRIGHTNESS value      */
#define VO_CAP_CONTRAST     0x00000080 /* driver can set CONTRAST value        */
#define VO_CAP_COLORKEY     0x00000100 /* driver can set COLORKEY value        */

#define ASPECT_AUTO        0
#define ASPECT_FULL        2 /* 4:3  */
#define ASPECT_ANAMORPHIC  1 /* 16:9 */
#define ASPECT_DVB         3 /* 1:2 */

vo_driver_t *init_video_out_mga () ;
vo_driver_t *init_video_out_xv () ;
vo_driver_t *init_video_out_xshm () ;

/* FIXME: Write output drivers for those too 
   vo_functions_t *init_video_out_x11 () ;
   FIXME
*/

#endif
