/* 
 * Copyright (C) 2000 the xine project
 * 
 * This file is part of xine, a unix video player.
 * 
 * xine is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * xine is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 * $Id: xine.c,v 1.96 2001/05/06 14:34:36 guenter Exp $
 *
 * top-level xine functions
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <dirent.h>
#include <dlfcn.h>
#include <Imlib.h>

#if defined (__linux__)
#include <endian.h>
#elif defined (__FreeBSD__)
#include <machine/endian.h>
#endif

#include "xine.h"
#include "audio_out/audio_out.h"
#include "video_out/video_out.h"
#include "demuxers/demux.h"
#include "buffer.h"
#include "libac3/ac3.h"
#include "libmpg123/mpg123.h"
#include "libmpg123/mpglib.h"
#include "libmpeg2/mpeg2.h"
#ifdef ARCH_X86
#include "libw32dll/w32codec.h"
#endif
#include "libspudec/spudec.h"
#include "input/input_plugin.h"
#include "metronom.h"
#include "configfile.h"
#include "monitor.h"

#ifdef HAVE_DXR3
#include <libdxr3/api.h>
extern dxr3_state_t state;
extern int output_spdif(uint8_t *data_start, uint8_t *data_end, int fd);
static uint32_t apts = 0, vpts = 0;
#endif

#ifdef HAVE_RSUB
#include <rsub.h>
#endif

extern uint32_t xine_debug;
extern const char *ao_available[];

#ifdef HAVE_DXR3
extern int tv_mode;
extern int overlay_mode;
extern int aspect_ratio;
#endif

#ifdef HAVE_RSUB
extern int enable_rsub;
extern char* xine_rsub_script;
#endif

/*
 * global variables - keep them to a minimum.
 */

Display        *gDisplay;
pthread_mutex_t gXLock;
ImlibData      *gImlib_data;
Window          gVideoWin;
ImlibImage     *gXineLogoImg;
Pixmap          gXineLogo;
int             gXineLogoWidth, gXineLogoHeight;

/* debugging purposes only */

#ifdef XLOCK_DEBUG
char * xlock_last_file;
int    xlock_last_line;
int    xlock_locked;
char * xunlock_last_file;
int    xunlock_last_line;
#endif

/*
#define TEST_FILE 
*/
#ifdef TEST_FILE
int             gTestFile=-1; 
#endif

typedef struct _xine_globals {
  input_plugin_t             mInputPlugins[INPUT_PLUGIN_MAX];
  int                        mNumInputPlugins;
  input_plugin_t            *mCurInput;

  demux_functions_t          mDemuxer[DEMUXER_PLUGIN_MAX];
  int                        mNumDemuxer;
  demux_functions_t         *mCurDemux;

  int                        mnStatus;
  off_t                      mnCurInputPos;
  char                       mCurMRL[1024];

  pthread_t                  mAudioThread;
  pthread_t                  mVideoThread;
  pthread_t                  mSPUThread;

  fifo_buffer_t             *mBufAudio;
  fifo_buffer_t             *mBufVideo;
  fifo_buffer_t             *mBufSPU;

  int                        mbVideoFinished;
  int                        mbAudioFinished;

  mpgaudio_t                *mpgaudio;

  int                        mnAudioChannel;
  int                        mnSPUChannel;

  clut_t                    *clut; /* array of 16 clut_t's */

  gui_status_callback_func_t mStatusCallback;

  /* ScreenSaver stuff */
  int                        timeout;
  int                        interval;
  int                        prefer_blanking;
  int                        allow_exposures;

  fifobuf_functions_t        *fifo_funcs;

  /* Lock for xine player functions */
  pthread_mutex_t            xine_lock;

} xine_globals_t ;

static xine_globals_t gXine;

#ifdef BIG_ENDIAN
static uint32_t default_palette[32] = {
  0x80801000, 0x80801000, 0x80808400, 0x8080eb00,
  0x80801000, 0x80801000, 0x80808400, 0x8080eb00,
  0x80801000, 0x80801000, 0x80808400, 0x8080eb00,
  0x80801000, 0x80801000, 0x80808400, 0x8080eb00,
};

#else

static uint32_t default_palette[32] = {
  0x00108080, 0x00108080, 0x00848080, 0x00eb8080,
  0x00108080, 0x00108080, 0x00848080, 0x00eb8080,
  0x00108080, 0x00108080, 0x00848080, 0x00eb8080,
  0x00108080, 0x00108080, 0x00848080, 0x00eb8080
};
#endif

#define LOCK_XINE { pthread_mutex_lock (&gXine.xine_lock); }
#define UNLOCK_XINE { pthread_mutex_unlock (&gXine.xine_lock); }

static clut_t* default_clut = (clut_t*) default_palette;

void xine_notify_stream_finished () {
  printf ("xine_notify_stream_finished\n");

  xine_stop ();

  gXine.mStatusCallback (gXine.mnStatus);
  XSync(gDisplay, False);
}

void *xine_audio_loop (void *dummy) {

  buf_element_t *pBuf;
  int bRunning = 1;

  while (bRunning) {

    pBuf = gXine.fifo_funcs->fifo_buffer_get (gXine.mBufAudio);

    if (gAudioOut) {

      if(pBuf->nType != BUF_MPEGELEMENT)
	gXine.mnCurInputPos = pBuf->nInputPos;

      if (gXine.mnStatus == XINE_PLAY) 
	gXine.mStatusCallback (gXine.mnStatus);

      switch (pBuf->nType) {
      case BUF_QUIT:
	bRunning = 0;
	break;
      case BUF_RESET:
	gXine.mbAudioFinished = 0;

	ac3_reset ();
	mpg_audio_reset (gXine.mpgaudio);
	
	break;
#ifdef ARCH_X86
      case BUF_AVIINIT:
	w32c_init_audio ((WAVEFORMATEX *)pBuf->pContent);
	
	break;
#endif
      
      case BUF_AC3AUDIO:

#ifdef HAVE_DXR3
	if (pBuf->nPTS != 0) {
	  if (dxr3_audio_set_pts(pBuf->nPTS))
	    fprintf(stderr, "audio setpts failed\n");
#if 0
	  fprintf(stderr, "apts:\t%i\n", pBuf->nPTS);
#endif
	}

#endif

	ac3_decode_data(pBuf->pContent, pBuf->pContent + pBuf->nSize,
			pBuf->nPTS);
    
	break;
	
      case BUF_MPEGAUDIO:

#ifdef HAVE_DXR3
	if (pBuf->nPTS != 0) {
	  apts = pBuf->nPTS;
#if 0
	  fprintf(stderr, "apts:\t%i\n", apts);
#endif
	  dxr3_audio_set_pts(pBuf->nPTS);
	}
#endif
	
	mpg_audio_decode_data(gXine.mpgaudio, pBuf->pContent, pBuf->pContent + pBuf->nSize,
			      pBuf->nPTS);
    
	break;
#ifdef ARCH_X86
      case BUF_MSAUDIO:
    
	w32c_decode_audio (pBuf->pContent, pBuf->nSize,
			   pBuf->bFrameEnd, pBuf->nPTS);
	
	break;
#endif

      case BUF_LINERPCM: 
	{
	  int i;
	  int16_t *p=(int16_t *)pBuf->pContent;
	  
	  gAudioOut->open (16, 48000, AO_MODE_STEREO);

#ifdef HAVE_DXR3
	  if (pBuf->nPTS != 0) dxr3_audio_set_pts(pBuf->nPTS);
#endif
	  
	  for ( i=0 ; i<pBuf->nSize/2 ; i++ ){
	    p[i]=ntohs( p[i] );
	  }
	  gAudioOut->write_audio_data ( p, pBuf->nSize/4, pBuf->nPTS);
	  break;
	}
	
      case BUF_STREAMEND:
	gXine.mbAudioFinished = 1;

	printf ("audio stream end\n");

	if (gXine.mbVideoFinished) {
	  xine_notify_stream_finished ();
	}
	break;
      }
    }
    gXine.fifo_funcs->buffer_pool_free (pBuf);
  }

  return NULL;

}

void *xine_spu_loop (void *dummy) {

  buf_element_t *pBuf;
  int bRunning = 1;

  while (bRunning) {

    pBuf = gXine.fifo_funcs->fifo_buffer_get (gXine.mBufSPU);

    switch (pBuf->nType) {
      case BUF_QUIT:
	bRunning = 0;
	break;
      case BUF_RESET:
	spudec_reset ();
#ifdef HAVE_DXR3
	dxr3_subpic_set_mode(DXR3_SPU_MODE_ON);
#endif
	break;
      case BUF_SPU:
#ifdef HAVE_DXR3
	if (pBuf->nPTS != 0) {
#if 0
	  fprintf(stderr, "spts:\t%i\n", pBuf->nPTS);
#endif
	  if (dxr3_subpic_set_pts(pBuf->nPTS))
	    fprintf(stderr, "spu setpts failed\n");
	}
	if (dxr3_subpic_write(pBuf->pContent,
			   pBuf->nSize) < 0)
	  fprintf(stderr, "spu write failed\n");
#else
        spudec_decode(pBuf->pContent, pBuf->nSize, pBuf->nPTS);
#endif
	break;        
    }
    gXine.fifo_funcs->buffer_pool_free (pBuf);
  }

  return NULL;

}

void *xine_video_loop () {

  buf_element_t *pBuf;
  int bRunning = 1;

  while (bRunning) {

    pBuf = gXine.fifo_funcs->fifo_buffer_get (gXine.mBufVideo);

    if (!gAudioOut || pBuf->nType == BUF_MPEGELEMENT) {
      gXine.mnCurInputPos = pBuf->nInputPos;
      if (gXine.mnStatus == XINE_PLAY)
	gXine.mStatusCallback (gXine.mnStatus);
    }

    switch (pBuf->nType) {
    case BUF_QUIT:
      bRunning = 0;
      break;

    case BUF_RESET:
      gXine.mbVideoFinished = 0;
#ifndef HAVE_DXR3
      mpeg2_reset ();
#endif
      break;
      
    case BUF_MPEG2VIDEO:
    case BUF_MPEG1VIDEO:
    case BUF_MPEGELEMENT:

#ifdef HAVE_RSUB
      if (enable_rsub && (pBuf->nPTS != 0))
	rsub_set_pts(pBuf->nPTS);
#endif

#ifdef HAVE_DXR3
      if (pBuf->nPTS != 0) {
	vpts = pBuf->nPTS;
#if 0
	fprintf(stderr, "vpts:\t%i\n", vpts);
#endif
	if (dxr3_video_set_pts(pBuf->nPTS))
	  fprintf(stderr, "video setpts failed\n");
    }

      if (dxr3_video_write(pBuf->pContent,
			   pBuf->nSize) < 0)
	fprintf(stderr, "video write failed\n");

#else

      mpeg2_decode_data(pBuf->pContent, 
			pBuf->pContent + pBuf->nSize, 
			pBuf->nPTS);

#endif

      break;
#ifdef ARCH_X86
    case BUF_AVIINIT:
      w32c_init_video ((BITMAPINFOHEADER *)pBuf->pContent);

      break;
    case BUF_AVIVIDEO:
    
      w32c_decode_video (pBuf->pContent, pBuf->nSize,
			 pBuf->bFrameEnd, pBuf->nPTS);

      break;
#endif
    case BUF_STREAMEND:
      gXine.mbVideoFinished = 1;
      printf ("video stream end\n");

#ifdef ARCH_X86
      w32c_close_audio ();
#endif

      if (gXine.mbAudioFinished || 
	  /* This is needed with stream without sound and Null audio driver */
	  (gXine.mbAudioFinished == 0 && gAudioOut == NULL)) { 
	xine_notify_stream_finished ();
      }
      break;
    }

    gXine.fifo_funcs->buffer_pool_free (pBuf);
  }

  return NULL;
}

static void xine_stop_locked () {

  if (!gXine.mCurInput) 
    return;

  gXine.mnStatus = XINE_STOP;

  if(gXine.mCurDemux) {
    gXine.mCurDemux->demux_stop ();
    gXine.mCurDemux = NULL;
  }

#ifdef HAVE_DXR3
  dxr3_set_playmode(DXR3_PLAYMODE_STOPPED);
#endif

#ifdef HAVE_RSUB
  if (enable_rsub)
    rsub_stop();
#endif

  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufVideo);
  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufAudio);
  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufSPU);

  if (gAudioOut)
    gAudioOut->close ();

  metronom_reset();
  metronom_stop_clock ();

#ifndef HAVE_DXR3
  vo_reset();
#endif

  gXine.mCurInput->close();
  gXine.mCurInput = NULL;
}

void xine_stop () {
  LOCK_XINE
  xine_stop_locked ();
  UNLOCK_XINE
}

static void xine_pause_locked ();

void xine_play_internal (char *MRL, int spos, off_t pos) {

  double     share ;
  off_t      len;
  int        i;
  video_info_t vinfo;

  xprintf (VERBOSE|LOOP, "xine open %s, start pos = %d\n",MRL, spos);
  printf ("xine open %s, start pos = %d\n",MRL, spos);

  if (gXine.mnStatus == XINE_PAUSE) {
    xine_pause_locked();
    return;
  }

  if (gXine.mnStatus != XINE_STOP) {
    xine_stop ();
  }

  /*
   * find demuxer
   */

  gXine.mCurDemux = NULL;
  for (i=0; i<gXine.mNumDemuxer; i++) {
    if (gXine.mDemuxer[i].demux_can_handle (MRL)) {
      gXine.mCurDemux = &gXine.mDemuxer[i];
      break;
    }
  }

  if (!gXine.mCurDemux) {
    printf ("error: couldn't find demuxer for >%s<\n",MRL);
    return;
  }

  /*
   * find input plugin
   */
   
  gXine.mCurInput = NULL;

  for (i = 0; i<gXine.mNumInputPlugins; i++) {
    if (gXine.mInputPlugins[i].open(MRL)) {

      gXine.mCurInput = &gXine.mInputPlugins[i];
#ifdef HAVE_RSUB
      if (enable_rsub) {
	rsub_set_chapter_mrl(MRL);
      }
#endif
      break;
    }
  }

  if (!gXine.mCurInput) {
    perror ("open input source");
    gXine.mCurDemux = NULL;
    return;
  }

#ifndef HAVE_DXR3  
  vo_set_logo_mode (0);
#endif

  /*
   * Init SPU decoder with colour lookup table. 
   */

  if(gXine.mCurInput->get_clut) {
    printf("Attempting to obtain CLUT\n");
    gXine.clut = gXine.mCurInput->get_clut();
  }

  if(gXine.clut == NULL) {
    printf("No CLUT found, using default (pretty ugly).\n");
    gXine.clut = default_clut;
  }
  spudec_init(gXine.clut);

#ifdef HAVE_DXR3
  dxr3_subpic_set_palette((char*) gXine.clut);
  dxr3_subpic_set_mode(DXR3_SPU_MODE_ON);
#endif

  if(gXine.mCurInput->get_video_info) {
    gXine.mCurInput->get_video_info(&vinfo);

#ifdef HAVE_RSUB
    if (enable_rsub) {
      rsub_set_fps(0, framerate_pts[vinfo.framerate], framerate_pts[vinfo.framerate]/2);
      printf("framerate: %u\n", vinfo.framerate);
    }
    rsub_start(gXine.mBufSPU);
#endif

#ifdef HAVE_DXR3
    {
      static int ratio_table[] = {DXR3_ASPECTRATIO_4_3, DXR3_ASPECTRATIO_4_3,
				  DXR3_ASPECTRATIO_4_3, DXR3_ASPECTRATIO_16_9};
      aspect_ratio = ratio_table[vinfo.ratio];
      dxr3_video_set_aspectratio(aspect_ratio);
      printf("ratio: %u\n", vinfo.ratio);
    }
#endif

  }

  /*
   * metronom
   */

  metronom_reset();

  /*
   * start demuxer
   */
  
  if (spos) {
    len = gXine.mCurInput->get_length ();
    share = (double) spos / 65535;
    pos = (off_t) (share * len) ;
  }

  gXine.mCurDemux->demux_select_audio_channel (gXine.mnAudioChannel);
  gXine.mCurDemux->demux_select_spu_channel (gXine.mnSPUChannel);
  gXine.mCurDemux->demux_start (gXine.mCurInput, gXine.mBufVideo,
				gXine.mBufAudio, gXine.mBufSPU, pos);

  gXine.mnStatus = XINE_PLAY;
  gXine.mnCurInputPos = pos;

#ifdef HAVE_DXR3
  dxr3_set_playmode(DXR3_PLAYMODE_PLAY);
#endif

  /*
   * remember MRL
   */

  strncpy (gXine.mCurMRL, MRL, 1024);

  /*
   * start clock
   */

  metronom_start_clock (0);
}

void xine_play (char *MRL, int spos) {
  LOCK_XINE
   if (gXine.mnStatus != XINE_PLAY)    
     xine_play_internal (MRL, spos, (off_t) 0);
  UNLOCK_XINE
}

static int xine_eject_locked(char *MRL) {
  int i;
  
  if(gXine.mCurInput == NULL) {
    
    for (i = 0; i<gXine.mNumInputPlugins; i++) {
      if (gXine.mInputPlugins[i].open(MRL)) {
	gXine.mCurInput = &gXine.mInputPlugins[i];
	gXine.mCurInput->close();
	break;
      }
    }
  }
  
  if (gXine.mnStatus == XINE_STOP
      && gXine.mCurInput && gXine.mCurInput->eject_media)
    return gXine.mCurInput->eject_media ();

  return 0;
}

int xine_eject(char *MRL) {
  int retval;
  LOCK_XINE
  retval = xine_eject_locked (MRL);
  UNLOCK_XINE
  return retval;
}

void xine_exit () {

  void *p;
  buf_element_t *pBuf;

  if (gXine.mCurDemux)
    gXine.mCurDemux->demux_stop ();

#ifdef HAVE_RSUB
  if (enable_rsub)
    rsub_stop();
#endif

  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufVideo);
  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufAudio);
  gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufSPU);

#ifdef HAVE_DXR3 /* unstick dxr3_write's */
  if (dxr3_get_status() == DXR3_STATUS_OPENED) {
    dxr3_set_playmode(DXR3_PLAYMODE_PLAY);
  }
#endif

    if (gAudioOut)
      gAudioOut->close ();

  pBuf = gXine.fifo_funcs->buffer_pool_alloc ();
  pBuf->nType = BUF_QUIT;
  gXine.fifo_funcs->fifo_buffer_put (gXine.mBufVideo, pBuf);
  pBuf = gXine.fifo_funcs->buffer_pool_alloc ();
  pBuf->nType = BUF_QUIT;
  gXine.fifo_funcs->fifo_buffer_put (gXine.mBufAudio, pBuf);
  pBuf = gXine.fifo_funcs->buffer_pool_alloc ();
  pBuf->nType = BUF_QUIT;
  gXine.fifo_funcs->fifo_buffer_put (gXine.mBufSPU, pBuf);

  pthread_join (gXine.mAudioThread, &p);
  pthread_join (gXine.mVideoThread, &p);
  pthread_join (gXine.mSPUThread, &p);

  gXine.mnStatus = XINE_QUIT;

#ifdef HAVE_DXR3
  if (dxr3_get_status() == DXR3_STATUS_OPENED) {
    dxr3_set_playmode(DXR3_PLAYMODE_STOPPED);
    dxr3_close();
  }
#endif

  /*
   * Restore screensaver parameters
   */
  XSetScreenSaver(gDisplay, gXine.timeout, gXine.interval, 
		  gXine.prefer_blanking, gXine.allow_exposures);

  config_file_save ();
}

static void xine_pause_locked () {

  if (gXine.mnStatus == XINE_PAUSE) {

    xprintf (VERBOSE, "xine play %s from %Ld\n",gXine.mCurMRL, gXine.mnCurInputPos);

    gXine.mnStatus = XINE_STOP;

    xine_play_internal (gXine.mCurMRL, 0, gXine.mnCurInputPos);
    /* gXine.mnPausePos = 0; */

  } else if (gXine.mnStatus == XINE_PLAY) {

    if (!gXine.mCurInput) 
      return;

    gXine.mnStatus = XINE_PAUSE;

    gXine.mCurDemux->demux_stop ();
    gXine.mCurDemux = NULL;

#ifdef HAVE_DXR3
    dxr3_set_playmode(DXR3_PLAYMODE_PAUSED);
#endif
    
    gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufVideo);
    gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufAudio);
    gXine.fifo_funcs->fifo_buffer_clear(gXine.mBufSPU);
    
    if (gAudioOut)
      gAudioOut->close ();

    metronom_reset();
    metronom_stop_clock ();

#ifndef HAVE_DXR3
    vo_reset ();
#endif

    gXine.mCurInput->close();
  }
}

void xine_pause () {
  LOCK_XINE
  xine_pause_locked ();
  UNLOCK_XINE
}


/*
 * update mCurMRL
 */
void xine_set_current_mrl( char* mrl ) {
  strncpy (gXine.mCurMRL, mrl, 1024);
}


void xine_load_demux_plugins (void) {
  DIR *dir;

  gXine.mDemuxer[0]    = *(init_demux_mpeg (gXine.fifo_funcs, xine_debug));
  gXine.mDemuxer[1]    = *(init_demux_mpeg_block (gXine.fifo_funcs, xine_debug));
  gXine.mDemuxer[2]    = *(init_demux_avi (gXine.fifo_funcs, xine_debug));
  gXine.mDemuxer[3]    = *(init_demux_mpeg_audio (gXine.fifo_funcs, xine_debug));
  gXine.mDemuxer[4]    = *(init_demux_mpeg_elem (gXine.fifo_funcs, xine_debug));
  gXine.mNumDemuxer    = 5;

  dir = opendir (XINE_DEMUXDIR) ;
  
  if (dir) {
    struct dirent *pEntry;
    
    while ((pEntry = readdir (dir)) != NULL) {
      char str[1024];
      void *plugin;
      
      int nLen = strlen (pEntry->d_name);
      
      if ((strncasecmp(pEntry->d_name, "demux_", 6) == 0) && 
	  ((pEntry->d_name[nLen-3]=='.') 
	   && (pEntry->d_name[nLen-2]=='s')
	   && (pEntry->d_name[nLen-1]=='o'))) {
	
	/*
	 * demux plugin found => load it
	 */
	
	sprintf (str, "%s/%s", XINE_DEMUXDIR, pEntry->d_name);
	
	if(!(plugin = dlopen (str, RTLD_LAZY))) {
	  fprintf(stderr, "%s(%d): %s doesn't seem to be installed (%s)\n", 
		  __FILE__, __LINE__, str, dlerror());
	  exit(1);
	}
	else {
	  
	  if (plugin) {
	    void *(*getinfo) (fifobuf_functions_t *, uint32_t);
	    
	    if((getinfo = dlsym(plugin, "demux_plugin_getinfo")) != NULL) {
	      demux_functions_t *dxp;
	      
	      dxp = (demux_functions_t *) getinfo(gXine.fifo_funcs, xine_debug);
	      dxp->handle = plugin;
	      dxp->filename = str;
	      gXine.mDemuxer[gXine.mNumDemuxer] = *dxp; 

	    
	      printf("demux plugin found : %s(%s)\n", 
		     gXine.mDemuxer[gXine.mNumDemuxer].filename,
		     pEntry->d_name);
	    
	      gXine.mNumDemuxer++;
	    }
	    
	    if(gXine.mNumDemuxer > DEMUXER_PLUGIN_MAX) {
	      fprintf(stderr, "%s(%d): too many demux plugins installed, "
		      "exiting.\n", __FILE__, __LINE__);
	      exit(1);
	    }
	  }
	}
      }
    }
  }

  if (gXine.mNumDemuxer == 5)
    printf ("No extra demux plugins found in %s\n", XINE_DEMUXDIR);
  
  /*
   * init demuxer
   */
  
  gXine.mCurDemux = NULL;
}

void xine_load_input_plugins () {
  DIR *dir;
  
  gXine.mNumInputPlugins = 0;
  
  dir = opendir (XINE_PLUGINDIR) ;
  
  if (dir) {
    struct dirent *pEntry;
    
    while ((pEntry = readdir (dir)) != NULL) {
      
      char str[1024];
      void *plugin;
      
      int nLen = strlen (pEntry->d_name);
      
      if ((strncasecmp(pEntry->d_name, "input_", 6) == 0) &&
	  ((pEntry->d_name[nLen-3]=='.') 
	  && (pEntry->d_name[nLen-2]=='s')
	  && (pEntry->d_name[nLen-1]=='o'))) {

	/*
	 * input plugin found => load it
	 */
	
	sprintf (str, "%s/%s", XINE_PLUGINDIR, pEntry->d_name);
	
	if(!(plugin = dlopen (str, RTLD_LAZY))) {
	  fprintf(stderr, "%s(%d): %s doesn't seem to be installed (%s)\n", 
		  __FILE__, __LINE__, str, dlerror());
	  exit(1);
	}
	else {
	  
	  if (plugin) {
	    void *(*getinfo) (uint32_t);
	    
	    if((getinfo = dlsym(plugin, "input_plugin_getinfo")) != NULL) {
	      input_plugin_t *ipp;
	      
	      ipp = (input_plugin_t *) getinfo(xine_debug);
	      ipp->handle = plugin;
	      ipp->filename = str;
	      gXine.mInputPlugins[gXine.mNumInputPlugins] = *ipp; 
	      
	      gXine.mInputPlugins[gXine.mNumInputPlugins].init();
	      
	      printf("input plugin found : %s(%s)\n", 
		     gXine.mInputPlugins[gXine.mNumInputPlugins].filename,
		     pEntry->d_name);
	      
	      gXine.mNumInputPlugins++;

	    }

	    if(gXine.mNumInputPlugins > INPUT_PLUGIN_MAX) {
	      fprintf(stderr, "%s(%d): too many input plugins installed, "
		      "exiting.\n", __FILE__, __LINE__);
	      exit(1);
	    }
	  }
	}
      }
    }
  }

  if (gXine.mNumInputPlugins == 0) {
    printf ("No input plugins found in %s! - "
	    "Did you install xine correctly??\n", XINE_PLUGINDIR);
    exit (1);
  }

}

void xine_init (int no_xv, gui_status_callback_func_t gui_status_callback, 
		int audio_driver_id, int ao_mode) {

  ac3_config_t    ac3_config;
  char            strLogo[256];

  gXine.mStatusCallback = gui_status_callback;

#ifdef TEST_FILE
  gTestFile = open ("/tmp/test.mp3", O_WRONLY | O_CREAT, 0644); 
#endif

  /*
   * init lock
   */

  pthread_mutex_init (&gXine.xine_lock, NULL);

  metronom_init ();

  /*
   * init imlib
   */

  gImlib_data   = Imlib_init(gDisplay);

  /*
   * load logo (IMPORTANT ;))
   */
  sprintf (strLogo, "%s/xine_logo.png", XINE_SKINDIR);
  gXineLogoImg = Imlib_load_image(gImlib_data, strLogo);
  
  if (!gXineLogoImg) {
    fprintf (stderr, "logo not found :-((\n");
    exit(1);
  }

  Imlib_render (gImlib_data, gXineLogoImg, 
		gXineLogoImg->rgb_width, gXineLogoImg->rgb_height);
  gXineLogo  = Imlib_move_image(gImlib_data, gXineLogoImg);
  gXineLogoWidth  = gXineLogoImg->rgb_width ;
  gXineLogoHeight = gXineLogoImg->rgb_height;

  /*
   * init dxr3
   */

#ifdef HAVE_DXR3
  if (dxr3_get_status() == DXR3_STATUS_CLOSED) {
    dxr3_open("/dev/em8300", "/etc/dxr3.ux");
  }
  if (overlay_mode == DXR3_OVERLAY_MODE_OFF) {
    dxr3_video_set_tvmode(tv_mode);
  }
  dxr3_video_set_aspectratio(aspect_ratio);
  dxr3_set_playmode(DXR3_PLAYMODE_STOPPED);
#endif

  /*
   * Init buffers
   */

  gXine.fifo_funcs = buffer_pool_init (2000, 4096);
  gXine.mBufVideo = gXine.fifo_funcs->fifo_buffer_new ();
  gXine.mBufAudio = gXine.fifo_funcs->fifo_buffer_new ();
  gXine.mBufSPU   = gXine.fifo_funcs->fifo_buffer_new ();

  xine_load_demux_plugins();

  /*
   * init rsub
   */

#ifdef HAVE_RSUB
  if (enable_rsub) {
    rsub_init(gXine.fifo_funcs, xine_rsub_script);
  }
#endif

  /*
   * init demuxer
   */
  
  gXine.mnAudioChannel = 0;
  gXine.mnSPUChannel   = 0;
  gXine.mnCurInputPos  = 0;

  printf ("xine_init: demuxer initialized\n");


  /*
   * init video output, start video thread
   */

#ifndef HAVE_DXR3
  vo_init (no_xv);
  mpeg2_init();
#endif
  mpeg2_reset();

  pthread_create (&gXine.mVideoThread, NULL, xine_video_loop, NULL) ;

  printf ("xine_init: video thread created\n");


  /*
   * init SPU decoder, start SPU thread
   */

  gXine.clut = default_clut;
  spudec_init(gXine.clut); 
#ifdef HAVE_DXR3
  dxr3_subpic_set_palette((char*) gXine.clut);
  dxr3_subpic_set_mode(DXR3_SPU_MODE_ON);
#endif

  pthread_create (&gXine.mSPUThread, NULL, xine_spu_loop, NULL) ;

  printf ("xine_init: SPU thread created\n");

  /*
   * init audio output, start audio thread
   */

  ac3_config.num_output_ch = 2;
  ac3_config.flags =  ao_mode;
  audio_out_init(audio_driver_id);

  if (!gAudioOut) {
    printf("audio driver '%s' init failed - no audio.\n", 
	   ao_available[audio_driver_id]);
  } else {
    ac3_init(&ac3_config, gAudioOut);
    gXine.mpgaudio = mpg_audio_init (gAudioOut);
  }

  /* Create thread only if audio is successfuly initialized. */
  if(gAudioOut)
    pthread_create (&gXine.mAudioThread, NULL, xine_audio_loop, NULL) ;
  
  printf ("xine_init: audio thread created\n");

  /*
   * load input plugins
   */

  xine_load_input_plugins ();

  printf ("xine_init: plugins loaded\n");

  /*
   * Store original screensaver parameters and sets new ones
   */
  XGetScreenSaver(gDisplay ,&gXine.timeout, &gXine.interval,
		  &gXine.prefer_blanking, &gXine.allow_exposures);
  XSetScreenSaver(gDisplay, 0, 0, DefaultBlanking, DefaultExposures);
  
}

int xine_get_audio_channel (void) {

  return gXine.mnAudioChannel;
}

void xine_select_audio_channel (int nChannel) {

  LOCK_XINE
  gXine.mnAudioChannel = nChannel;

  if (gXine.mCurDemux) {
    gXine.mCurDemux->demux_select_audio_channel (gXine.mnAudioChannel);
  }
  UNLOCK_XINE
}

int xine_get_spu_channel (void) {

  return gXine.mnSPUChannel;
}

void xine_select_spu_channel (int nChannel) {

  LOCK_XINE
  gXine.mnSPUChannel = (nChannel >= -1 ? nChannel : -1);

  if (gXine.mCurDemux) 
    gXine.mCurDemux->demux_select_spu_channel (gXine.mnSPUChannel);
  UNLOCK_XINE
}

input_plugin_t* xine_get_input_plugin_list (int *nInputPlugins) {

  *nInputPlugins = gXine.mNumInputPlugins;
  return gXine.mInputPlugins;
}

static int xine_get_current_position_locked (void) {

  /* off_t pos; */
  off_t len;
  double share;
  
  if (!gXine.mCurInput) {
    xprintf (VERBOSE|INPUT, "xine_get_current_position: no input source\n");
    return 0;
  }
  
  /* pos = gXine.mCurInput->seek (0, SEEK_CUR); */
  len = gXine.mCurInput->get_length ();

  share = (double) gXine.mnCurInputPos / (double) len * 65535;

  return (int) share;
}

int xine_get_current_position (void) {
  int retval;
  LOCK_XINE
  retval = xine_get_current_position_locked ();
  UNLOCK_XINE
  return retval;
}

int xine_get_status(void) {

  return gXine.mnStatus;
}
